# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified from Hydra
import copy
import functools
from enum import Enum
from textwrap import dedent
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union

from hydra._internal.utils import _locate
from hydra.errors import InstantiationException
from hydra.types import ConvertMode, TargetConf
from lightning_utilities.core.rank_zero import rank_zero_only
from omegaconf import OmegaConf, SCMode
from omegaconf._utils import is_structured_config
from rich import print
from rich.panel import Panel
from rich.syntax import Syntax

PRINT_FUNCTION_CALL = True
"""
Flag to determine whether function calls should be printed.
If set to True, function calls will be printed.
"""

PRINT_FUNCTION_CALL_FUNC = print
"""
Function to be used for printing function calls.
"""

CATCH_EXCEPTION = True


def _resolve_callable_name(f: Callable[..., Any]) -> str:
    # Get the module name
    module_name = f.__module__
    # Get the qualname (fully qualified name) of the callable
    qualname = f.__qualname__
    # If the module is not "__main__", include it in the full name
    if module_name != "__main__":
        full_name = f"{module_name}.{qualname}"
    else:
        full_name = qualname

    return full_name


def _format_args_kwargs(args, kwargs):
    result_strings = []
    if len(args) > 0:
        result_strings.append(", ".join(repr(arg) for arg in args))
    if len(kwargs) > 0:
        result_strings.append(", ".join(f"{k}={repr(v)}" for k, v in kwargs.items()))

    if len(result_strings) == 0:
        return ""
    else:
        return ", ".join(result_strings)


class _Keys(str, Enum):
    """Special keys in configs used by instantiate."""

    TARGET = "_target_"
    CONVERT = "_convert_"
    RECURSIVE = "_recursive_"
    ARGS = "_args_"
    PARTIAL = "_partial_"


def _is_target(x: Any) -> bool:
    if isinstance(x, dict):
        return "_target_" in x
    if OmegaConf.is_dict(x):
        return "_target_" in x
    return False


def _extract_pos_args(input_args: Any, kwargs: Any) -> Tuple[Any, Any]:
    config_args = kwargs.pop(_Keys.ARGS, ())
    output_args = config_args

    if isinstance(config_args, Sequence):
        if len(input_args) > 0:
            output_args = input_args
    else:
        raise InstantiationException(
            f"Unsupported _args_ type: '{type(config_args).__name__}'. value: '{config_args}'"
        )

    return output_args, kwargs


def _call_target(
    _target_: Callable[..., Any],
    _partial_: bool,
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    full_key: str,
) -> Any:
    """Call target (type) with args and kwargs."""
    try:
        args, kwargs = _extract_pos_args(args, kwargs)
        # detaching configs from parent.
        # At this time, everything is resolved and the parent link can cause
        # issues when serializing objects in some scenarios.
        for arg in args:
            if OmegaConf.is_config(arg):
                arg._set_parent(None)
        for v in kwargs.values():
            if OmegaConf.is_config(v):
                v._set_parent(None)
    except Exception as e:
        msg = (
            f"Error in collecting args and kwargs for '{_convert_target_to_string(_target_)}':"
            + f"\n{repr(e)}"
        )
        if full_key:
            msg += f"\nfull_key: {full_key}"

        raise InstantiationException(msg) from e

    if _partial_:
        if PRINT_FUNCTION_CALL and getattr(rank_zero_only, "rank", 0) == 0:
            call_str = f"functools.partial({_resolve_callable_name(_target_)}, {_format_args_kwargs(args, kwargs)})"
            PRINT_FUNCTION_CALL_FUNC(
                Panel(
                    Syntax(call_str, "python", theme="monokai", word_wrap=True),
                    title="Instantiate by calling partial",
                    border_style="cyan",
                )
            )

        if CATCH_EXCEPTION:
            try:
                return functools.partial(_target_, *args, **kwargs)
            except Exception as e:
                msg = (
                    f"Error in creating partial({_convert_target_to_string(_target_)}, ...) object:"
                    + f"\n{repr(e)}"
                )
                if full_key:
                    msg += f"\nfull_key: {full_key}"
                raise InstantiationException(msg) from e
        else:
            return functools.partial(_target_, *args, **kwargs)
    else:
        if PRINT_FUNCTION_CALL and getattr(rank_zero_only, "rank", 0) == 0:
            call_str = f"{_resolve_callable_name(_target_)}({_format_args_kwargs(args, kwargs)})"
            PRINT_FUNCTION_CALL_FUNC(
                Panel(
                    Syntax(call_str, "python", theme="monokai", word_wrap=True),
                    title="Instantiate by calling function",
                    border_style="green",
                )
            )
        if CATCH_EXCEPTION:
            try:
                return _target_(*args, **kwargs)
            except Exception as e:
                msg = f"Error in call to target '{_convert_target_to_string(_target_)}':\n{repr(e)}"
                if full_key:
                    msg += f"\nfull_key: {full_key}"
                raise InstantiationException(msg) from e
        else:
            return _target_(*args, **kwargs)


def _convert_target_to_string(t: Any) -> Any:
    if callable(t):
        return f"{t.__module__}.{t.__qualname__}"
    else:
        return t


def _prepare_input_dict_or_list(d: Union[Dict[Any, Any], List[Any]]) -> Any:
    res: Any
    if isinstance(d, dict):
        res = {}
        for k, v in d.items():
            if k == "_target_":
                v = _convert_target_to_string(d["_target_"])
            elif isinstance(v, (dict, list)):
                v = _prepare_input_dict_or_list(v)
            res[k] = v
    elif isinstance(d, list):
        res = []
        for v in d:
            if isinstance(v, (list, dict)):
                v = _prepare_input_dict_or_list(v)
            res.append(v)
    else:
        assert False
    return res


def _resolve_target(
    target: Union[str, type, Callable[..., Any]], full_key: str
) -> Union[type, Callable[..., Any]]:
    """Resolve target string, type or callable into type or callable."""
    if isinstance(target, str):
        try:
            target = _locate(target)
        except Exception as e:
            msg = f"Error locating target '{target}', set env var HYDRA_FULL_ERROR=1 to see chained exception."
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise InstantiationException(msg) from e
    if not callable(target):
        msg = f"Expected a callable target, got '{target}' of type '{type(target).__name__}'"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise InstantiationException(msg)
    return target


def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
    """
    :param config: An config object describing what to call and what params to use.
                   In addition to the parameters, the config must contain:
                   _target_ : target class or callable name (str)
                   And may contain:
                   _args_: List-like of positional arguments to pass to the target
                   _recursive_: Construct nested objects as well (bool).
                                True by default.
                                may be overridden via a _recursive_ key in
                                the kwargs
                   _convert_: Conversion strategy
                        none    : Passed objects are DictConfig and ListConfig, default
                        partial : Passed objects are converted to dict and list, with
                                  the exception of Structured Configs (and their fields).
                        object  : Passed objects are converted to dict and list.
                                  Structured Configs are converted to instances of the
                                  backing dataclass / attr class.
                        all     : Passed objects are dicts, lists and primitives without
                                  a trace of OmegaConf containers. Structured configs
                                  are converted to dicts / lists too.
                   _partial_: If True, return functools.partial wrapped method or object
                              False by default. Configure per target.
    :param args: Optional positional parameters pass-through
    :param kwargs: Optional named parameters to override
                   parameters in the config object. Parameters not present
                   in the config objects are being passed as is to the target.
                   IMPORTANT: dataclasses instances in kwargs are interpreted as config
                              and cannot be used as passthrough
    :return: if _target_ is a class name: the instantiated object
             if _target_ is a callable: the return value of the call
    """

    # Return None if config is None
    if config is None:
        return None

    # TargetConf edge case
    if isinstance(config, TargetConf) and config._target_ == "???":
        # Specific check to give a good warning about failure to annotate _target_ as a string.
        raise InstantiationException(
            dedent(
                f"""\
                Config has missing value for key `_target_`, cannot instantiate.
                Config type: {type(config).__name__}
                Check that the `_target_` key in your dataclass is properly annotated and overridden.
                A common problem is forgetting to annotate _target_ as a string : '_target_: str = ...'"""
            )
        )
        # TODO: print full key

    if isinstance(config, (dict, list)):
        config = _prepare_input_dict_or_list(config)

    kwargs = _prepare_input_dict_or_list(kwargs)

    # Structured Config always converted first to OmegaConf
    if is_structured_config(config) or isinstance(config, (dict, list)):
        config = OmegaConf.structured(config, flags={"allow_objects": True})

    if OmegaConf.is_dict(config):
        # Finalize config (convert targets to strings, merge with kwargs)
        config_copy = copy.deepcopy(config)
        config_copy._set_flag(
            flags=["allow_objects", "struct", "readonly"], values=[True, False, False]
        )
        config_copy._set_parent(config._get_parent())
        config = config_copy

        if kwargs:
            config = OmegaConf.merge(config, kwargs)

        OmegaConf.resolve(config)

        _recursive_ = config.pop(_Keys.RECURSIVE, True)
        _convert_ = config.pop(_Keys.CONVERT, ConvertMode.NONE)
        _partial_ = config.pop(_Keys.PARTIAL, False)

        return instantiate_node(
            config, *args, recursive=_recursive_, convert=_convert_, partial=_partial_
        )
    elif OmegaConf.is_list(config):
        # Finalize config (convert targets to strings, merge with kwargs)
        config_copy = copy.deepcopy(config)
        config_copy._set_flag(
            flags=["allow_objects", "struct", "readonly"], values=[True, False, False]
        )
        config_copy._set_parent(config._get_parent())
        config = config_copy

        OmegaConf.resolve(config)

        _recursive_ = kwargs.pop(_Keys.RECURSIVE, True)
        _convert_ = kwargs.pop(_Keys.CONVERT, ConvertMode.NONE)
        _partial_ = kwargs.pop(_Keys.PARTIAL, False)

        if _partial_:
            raise InstantiationException(
                "The _partial_ keyword is not compatible with top-level list instantiation"
            )

        return instantiate_node(
            config, *args, recursive=_recursive_, convert=_convert_, partial=_partial_
        )
    else:
        raise InstantiationException(
            dedent(
                f"""\
                Cannot instantiate config of type {type(config).__name__}.
                Top level config must be an OmegaConf DictConfig/ListConfig object,
                a plain dict/list, or a Structured Config class or instance."""
            )
        )


def _convert_node(node: Any, convert: Union[ConvertMode, str]) -> Any:
    if OmegaConf.is_config(node):
        if convert == ConvertMode.ALL:
            node = OmegaConf.to_container(node, resolve=True)
        elif convert == ConvertMode.PARTIAL:
            node = OmegaConf.to_container(
                node, resolve=True, structured_config_mode=SCMode.DICT_CONFIG
            )
        elif convert == ConvertMode.OBJECT:
            node = OmegaConf.to_container(
                node, resolve=True, structured_config_mode=SCMode.INSTANTIATE
            )
    return node


def instantiate_node(
    node: Any,
    *args: Any,
    convert: Union[str, ConvertMode] = ConvertMode.NONE,
    recursive: bool = True,
    partial: bool = False,
) -> Any:
    # Return None if config is None
    if node is None or (OmegaConf.is_config(node) and node._is_none()):
        return None

    if not OmegaConf.is_config(node):
        return node

    # Override parent modes from config if specified
    if OmegaConf.is_dict(node):
        # using getitem instead of get(key, default) because OmegaConf will raise an exception
        # if the key type is incompatible on get.
        convert = node[_Keys.CONVERT] if _Keys.CONVERT in node else convert
        recursive = node[_Keys.RECURSIVE] if _Keys.RECURSIVE in node else recursive
        partial = node[_Keys.PARTIAL] if _Keys.PARTIAL in node else partial

    full_key = node._get_full_key(None)

    if not isinstance(recursive, bool):
        msg = f"Instantiation: _recursive_ flag must be a bool, got {type(recursive)}"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise TypeError(msg)

    if not isinstance(partial, bool):
        msg = f"Instantiation: _partial_ flag must be a bool, got {type(partial)}"
        if node and full_key:
            msg += f"\nfull_key: {full_key}"
        raise TypeError(msg)

    # If OmegaConf list, create new list of instances if recursive
    if OmegaConf.is_list(node):
        items = [
            instantiate_node(item, convert=convert, recursive=recursive)
            for item in node._iter_ex(resolve=True)
        ]

        if convert in (ConvertMode.ALL, ConvertMode.PARTIAL, ConvertMode.OBJECT):
            # If ALL or PARTIAL or OBJECT, use plain list as container
            return items
        else:
            # Otherwise, use ListConfig as container
            lst = OmegaConf.create(items, flags={"allow_objects": True})
            lst._set_parent(node)
            return lst

    elif OmegaConf.is_dict(node):
        exclude_keys = set({"_target_", "_convert_", "_recursive_", "_partial_"})
        if _is_target(node):
            _target_ = _resolve_target(node.get(_Keys.TARGET), full_key)
            kwargs = {}
            is_partial = node.get("_partial_", False) or partial
            for key in node.keys():
                if key not in exclude_keys:
                    if OmegaConf.is_missing(node, key) and is_partial:
                        continue
                    value = node[key]
                    if recursive:
                        value = instantiate_node(
                            value, convert=convert, recursive=recursive
                        )
                    kwargs[key] = _convert_node(value, convert)

            return _call_target(_target_, partial, args, kwargs, full_key)
        else:
            # If ALL or PARTIAL non structured or OBJECT non structured,
            # instantiate in dict and resolve interpolations eagerly.
            if convert == ConvertMode.ALL or (
                convert in (ConvertMode.PARTIAL, ConvertMode.OBJECT)
                and node._metadata.object_type in (None, dict)
            ):
                dict_items = {}
                for key, value in node.items():
                    # list items inherits recursive flag from the containing dict.
                    dict_items[key] = instantiate_node(
                        value, convert=convert, recursive=recursive
                    )
                return dict_items
            else:
                # Otherwise use DictConfig and resolve interpolations lazily.
                cfg = OmegaConf.create({}, flags={"allow_objects": True})
                for key, value in node.items():
                    cfg[key] = instantiate_node(
                        value, convert=convert, recursive=recursive
                    )
                cfg._set_parent(node)
                cfg._metadata.object_type = node._metadata.object_type
                if convert == ConvertMode.OBJECT:
                    return OmegaConf.to_object(cfg)
                return cfg

    else:
        assert False, f"Unexpected config type : {type(node).__name__}"



import functools
import multiprocessing
import os
import sys
from multiprocessing import Pool
from typing import Any, Dict, List, Literal, Optional

import numpy as np
from datasets import Dataset, DatasetDict, load_dataset, load_from_disk
from lightning.fabric.utilities import rank_zero_only
from tqdm.auto import tqdm
from typing_extensions import TYPE_CHECKING

import fusion_bench

from .arc import Example, Task
from .preprocess import get_augmenters, process_task, process_task_for_ttt

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer


def _get_formatter(format: Literal["new", "barc"]):
    from . import messagers, representers

    if format == "new":
        standard_formatter = representers.TextTaskRepresenter(
            example_representer=representers.TextExampleRepresenter(
                io_sep=" -> ",
                input_header="",
                output_header="",
                output_footer="#",
                grid_representer=representers.PythonListGridRepresenter(),
            )
        )
        formatter = messagers.GPTTextMessageRepresenterV2(
            task_representer=standard_formatter
        )
    elif format == "barc":
        formatter = messagers.GPTTextMessageRepresenterForBarc(
            prompt=(
                "Cutting Knowledge Date: December 2023\n"
                "Today Date: 26 Jul 2024\n\n"
                "You are a world-class puzzle solver with exceptional pattern recognition skills. "
                "Your task is to analyze puzzles, spot patterns, and provide direct solutions."
            ),
            task_representer=representers.TextTaskRepresenter(
                example_representer=representers.TextExampleRepresenter(
                    grid_representer=representers.WordGridRepresenter(),
                    input_header="Input:\n",
                    output_header="\nOutput:\n",
                    io_sep="\n",
                )
            ),
        )
    else:
        formatter = messagers.GPTTextMessageRepresenterV2()

    return formatter


def _join_list(lists: List[List[Any]]) -> List[Any]:
    ans = []
    for l in lists:
        ans.extend(l)
    return ans


def _to_tasks(
    train_data: List[Dict[str, Any]],
    test_data: List[Dict[str, Any]],
    name: str,
) -> List[Task]:
    # train examples
    train_examples = [
        Example(input=np.asarray(data["input"]), output=np.asarray(data["output"]))
        for data in train_data
    ]
    # test example
    test_examples = [
        Example(input=np.asarray(data["input"]), output=np.asarray(data["output"]))
        for data in test_data
    ]
    tasks = [
        Task(train_examples=train_examples, test_example=test_example, name=name)
        for test_example in test_examples
    ]
    return tasks


def tokenizer_tasks_for_ttt(
    tasks: List[Task],
    tokenizer: "PreTrainedTokenizer",
    use_data_augmentation: bool = True,
    permute_n: int = 1,
    seed: int = 0,
):
    if not use_data_augmentation:
        augmenters_to_apply = []
    else:
        augmenters_to_apply = get_augmenters(
            include_basic=True,
            include_size=True,
            include_chain=True,
            include_repeat=True,
        )

    formatter = _get_formatter("new")
    processor = functools.partial(
        process_task_for_ttt,
        augmenters=augmenters_to_apply,
        formatter=formatter,
        tokenizer=tokenizer,
        permute_n=permute_n,
        Nmax=250,
        seed=seed,
    )

    # with Pool(multiprocessing.cpu_count()) as p:
    #     data = p.map(processor, tasks)
    data = _join_list(
        [
            processor(task)
            for task in tqdm(
                tasks,
                desc="Processing tasks",
                dynamic_ncols=True,
                leave=False,
                disable=not rank_zero_only.rank == 0,
            )
        ]
    )
    dataset = Dataset.from_list(data)
    return dataset


def tokenizer_tasks(
    tasks: List[Task],
    tokenizer: "PreTrainedTokenizer",
):
    formatter = _get_formatter("new")
    processor = functools.partial(
        process_task, formatter=formatter, tokenizer=tokenizer
    )

    # with Pool(multiprocessing.cpu_count()) as p:
    #     data = p.map(processor, tasks)
    data = _join_list(
        [
            processor(task)
            for task in tqdm(
                tasks,
                desc="Processing tasks",
                dynamic_ncols=True,
                leave=False,
                disable=not rank_zero_only.rank == 0,
            )
        ]
    )
    dataset = Dataset.from_list(data)
    return dataset


def load_tokenized_arc_agi_dataset_for_ttt(
    tokenizer: Optional["PreTrainedTokenizer"],
    path: str = "dataartist/arc-agi",
    split: Optional[str] = None,
    cache_path: Optional[str] = None,
    use_data_augmentation: bool = True,
    permute_n: int = 1,
    seed: int = 0,
    max_num_tasks: Optional[int] = None,
):
    # regularize split
    split = split.lower() if split is not None else split

    # load cached dataset if available
    if cache_path is not None and fusion_bench.utils.path.path_is_dir_and_not_empty(
        cache_path
    ):
        datasets = load_from_disk(cache_path)
        if split is None and split in datasets.column_names:
            return datasets[split]
        else:
            return datasets
    else:
        assert (
            tokenizer is not None
        ), "Cached dataset not found. Need tokenizer to process the raw data."

    # load raw dataset
    datasets = load_dataset(path)
    datasets = DatasetDict(
        {"train": datasets["training"], "test": datasets["evaluation"]}
    )
    if split is None:
        converted_datasets: Dict[str, List[Task]] = {
            "train": _join_list(
                [
                    _to_tasks(
                        task["train"],
                        task["test"],
                        task["id"],
                    )
                    for task in datasets["train"]
                ]
            ),
            "test": _join_list(
                [
                    _to_tasks(
                        task["train"],
                        task["test"],
                        task["id"],
                    )
                    for task in datasets["test"]
                ]
            ),
        }
        if max_num_tasks is not None:
            # limit the number of tasks, useful for debugging
            converted_datasets = {
                split: converted_datasets[split][:max_num_tasks]
                for split in converted_datasets
            }
        converted_datasets = {
            split: tokenizer_tasks_for_ttt(
                converted_datasets[split],
                tokenizer,
                use_data_augmentation,
                permute_n,
                seed,
            )
            for split in tqdm(
                converted_datasets,
                desc="Processing splits",
                dynamic_ncols=True,
                disable=not rank_zero_only.rank == 0,
            )
        }
        converted_datasets = DatasetDict(converted_datasets)
    else:  # split is not None
        converted_datasets = _join_list(
            [
                _to_tasks(
                    task["train"],
                    task["test"],
                    task["id"],
                )
                for task in datasets[split]
            ]
        )
        if max_num_tasks is not None:
            # limit the number of tasks, useful for debugging
            converted_datasets = converted_datasets[:max_num_tasks]
        converted_datasets = tokenizer_tasks_for_ttt(
            converted_datasets, tokenizer, use_data_augmentation, permute_n, seed
        )

    if cache_path is not None and rank_zero_only.rank == 0:
        os.makedirs(cache_path, exist_ok=True)
        converted_datasets.save_to_disk(cache_path)
    return converted_datasets


def load_tokenized_arc_agi_dataset(
    tokenizer: Optional["PreTrainedTokenizer"],
    path: str = "dataartist/arc-agi",
    split: Optional[str] = None,
    cache_path: Optional[str] = None,
    max_num_tasks: Optional[int] = None,
):
    """
    Loads and tokenizes the ARC-AGI dataset.

    Args:
        tokenizer (Optional[PreTrainedTokenizer]): The tokenizer to use for tokenizing the dataset.
        path (str, optional): The path to the dataset. Defaults to "dataartist/arc-agi".
        split (Optional[str], optional): The dataset split to load (e.g., "train", "test"). Defaults to None.
        cache_path (Optional[str], optional): The path to cache the processed dataset. Defaults to None.
        max_num_tasks (Optional[int], optional): The maximum number of tasks to load. Useful for debugging. Defaults to None.

    Returns:
        DatasetDict or Dataset: The tokenized dataset, either as a DatasetDict if split is None, or as a Dataset if a specific split is specified.
    """
    # regularize split
    split = split.lower() if split is not None else split

    # load cached dataset if available
    if cache_path is not None and fusion_bench.utils.path.path_is_dir_and_not_empty(
        cache_path
    ):
        datasets = load_from_disk(cache_path)
        if split is None and split in datasets.column_names:
            return datasets[split]
        else:
            return datasets
    else:
        assert (
            tokenizer is not None
        ), "Cached dataset not found. Need tokenizer to process the raw data."

    # load raw dataset
    datasets = load_dataset(path)
    datasets = DatasetDict(
        {"train": datasets["training"], "test": datasets["evaluation"]}
    )
    if split is None:
        converted_datasets: Dict[str, List[Task]] = {
            "train": _join_list(
                [
                    _to_tasks(
                        task["train"],
                        task["test"],
                        task["id"],
                    )
                    for task in datasets["train"]
                ]
            ),
            "test": _join_list(
                [
                    _to_tasks(
                        task["train"],
                        task["test"],
                        task["id"],
                    )
                    for task in datasets["test"]
                ]
            ),
        }
        if max_num_tasks is not None:
            # limit the number of tasks, useful for debugging
            converted_datasets = {
                split: converted_datasets[split][:max_num_tasks]
                for split in converted_datasets
            }
        converted_datasets = {
            split: tokenizer_tasks(converted_datasets[split], tokenizer)
            for split in tqdm(
                converted_datasets,
                desc="Processing splits",
                dynamic_ncols=True,
                disable=not rank_zero_only.rank == 0,
            )
        }
        converted_datasets = DatasetDict(converted_datasets)
    else:  # split is not None
        converted_datasets = _join_list(
            [
                _to_tasks(
                    task["train"],
                    task["test"],
                    task["id"],
                )
                for task in datasets[split]
            ]
        )
        if max_num_tasks is not None:
            # limit the number of tasks, useful for debugging
            converted_datasets = converted_datasets[:max_num_tasks]
        converted_datasets = tokenizer_tasks(converted_datasets, tokenizer)

    if cache_path is not None and rank_zero_only.rank == 0:
        os.makedirs(cache_path, exist_ok=True)
        converted_datasets.save_to_disk(cache_path)
    return converted_datasets



import logging
from typing import Dict, List, Tuple  # noqa: F401

import torch
import torch.nn.functional as F
from torch import Tensor, nn

log = logging.getLogger(__name__)


class ExpertNotTrainedError(Exception):
    pass


def _is_all_zeros(tensor: Tensor | List[Tensor]) -> bool:
    if isinstance(tensor, Tensor):
        return torch.allclose(tensor, torch.zeros_like(tensor))
    else:
        return all(_is_all_zeros(t) for t in tensor)


def _svd(w: Tensor, full_matrices=True) -> Tuple[Tensor, Tensor, Tensor]:
    u, s, vh = torch.linalg.svd(
        w, full_matrices=full_matrices, driver="gesvd" if w.is_cuda else None
    )
    v = vh.T
    return u, s, v


def svd(
    w: Tensor, full_matrices=True, accelerator=None
) -> Tuple[Tensor, Tensor, Tensor]:
    if accelerator is None:
        return _svd(w, full_matrices=full_matrices)
    original_device = w.device
    w = w.to(accelerator)
    u, s, v = _svd(w)
    return u.to(original_device), s.to(original_device), v.to(original_device)


class SmileGate(nn.Module):
    def __init__(
        self,
        input_features: int,
        w_diff_list: List[Tensor],
        k: int,
        svd_list=None,  # cached `svd_list`, pass it to avoid recomputing
        upscaling_accelerator=None,
    ):
        super().__init__()
        self.input_features = input_features
        self.num_experts = len(w_diff_list)
        weights = []
        for i, w_diff in enumerate(w_diff_list):
            if svd_list is None:
                u, s, v = svd(w_diff, accelerator=upscaling_accelerator)
            else:
                u, s, v = svd_list[i]
            u = u[:, :k]
            s = s[:k]
            v = v[:, :k]

            # weights.append((s * v).T)
            weights.append(v.T)
        self.k = s.size(0)  # k is the actual k after truncation

        weights = (
            torch.stack(weights, dim=0)
            .reshape(self.num_experts * self.k, -1)
            .contiguous()
        )
        self.weights = nn.Parameter(
            weights
        )  # weights should be a tensor of shape (num_experts * k, n)

    def forward(self, x: Tensor):
        batch_size = x.size(0)
        if self.num_experts == 1:
            return torch.ones(batch_size, 1, device=x.device, dtype=x.dtype)

        routing_weights = F.linear(x, self.weights).view(
            batch_size, self.num_experts, self.k
        )
        routing_weights = routing_weights.norm(p=2, dim=2)
        return routing_weights


class SmileCompressedLinear(nn.Module):
    def __init__(self, model: nn.Linear, k: int, svd_cache=None):
        super().__init__()
        if svd_cache is None:
            u, s, v = svd(model.weight)
        else:
            u, s, v = svd_cache
        if k > 0:
            u = u[:, :k]
            s = s[:k]
            v = v[:, :k]

        self.u = nn.Parameter(u)
        self.svh = nn.Parameter((s * v).T)

        if model.bias is not None:
            self.bias = nn.Parameter(model.bias.data, requires_grad=True)
        else:
            self.register_parameter("bias", None)

    def forward(self, x):
        x = F.linear(x, self.svh)
        x = F.linear(x, self.u, self.bias)
        return x


class SmileMoELinear(nn.Module):
    @torch.no_grad()
    def __init__(
        self,
        pretrained_model: nn.Linear,
        finetuned_models: List[nn.Linear],
        gate_k: int,
        k: int,
        top_k: int = 1,
        full_matrices=True,
        upscaling_accelerator=None,
        routing_use_diff=True,
    ):
        super().__init__()
        self.num_experts = len(finetuned_models)
        self.top_k = top_k
        self.k = k
        self.gate_k = gate_k
        self.in_features = pretrained_model.in_features
        self.out_features = pretrained_model.out_features

        w_diff_list = [m.weight - pretrained_model.weight for m in finetuned_models]
        if _is_all_zeros(w_diff_list):
            # All fine-tuned models are identical to the pretrained model
            raise ExpertNotTrainedError()

        if routing_use_diff or k > 0:
            svd_cache_list = [
                svd(w, full_matrices=full_matrices, accelerator=upscaling_accelerator)
                for w in w_diff_list
            ]  # the svd cache list to avoid recomputing

        # construct the gate network
        if routing_use_diff:
            self.gate = SmileGate(
                input_features=self.in_features,
                w_diff_list=w_diff_list,
                k=gate_k,
                svd_list=svd_cache_list,
                upscaling_accelerator=upscaling_accelerator,
            )
        else:
            self.gate = SmileGate(
                input_features=self.in_features,
                w_diff_list=[m.weight for m in finetuned_models],
                k=gate_k,
                svd_list=None,
                upscaling_accelerator=upscaling_accelerator,
            )

        # construct experts
        for m, w_diff in zip(finetuned_models, w_diff_list):
            m.weight.data = w_diff
        if k > 0:
            experts = [
                SmileCompressedLinear(m, k, svd_cache=svd_cache)
                for m, svd_cache in zip(finetuned_models, svd_cache_list)
            ]
        else:
            # if k is not set (<0), we use the full fine-tuned model
            experts = finetuned_models
        self.experts = nn.ModuleList(experts)

        if pretrained_model.bias is not None:
            for m in experts:
                m.bias.data = m.bias.data - pretrained_model.bias
        # assign the pretrained model (the shared part)
        self.pretrained_model = pretrained_model

    def forward(self, hidden_states: Tensor):
        pretrained_out = self.pretrained_model(hidden_states)

        input_shape = hidden_states.size()
        hidden_states = hidden_states.view(-1, self.in_features)

        router_logits = self.gate(hidden_states)
        routing_weights = F.softmax(router_logits, dim=1)
        # sample the expert according to the routing weights
        routing_weights, selected_experts = torch.topk(
            routing_weights, self.top_k, dim=-1
        )
        routing_weights /= routing_weights.sum(dim=-1, keepdim=True)

        final_hidden_states = torch.zeros(
            (hidden_states.size(0), self.out_features),
            dtype=hidden_states.dtype,
            device=hidden_states.device,
        )

        # One hot encode the selected experts to create an expert mask
        # this will be used to easily index which expert is going to be sollicitated
        expert_mask = torch.nn.functional.one_hot(
            selected_experts, num_classes=self.num_experts
        ).permute(2, 1, 0)

        # Loop over all available experts in the model and perform the computation on each expert
        for expert_idx in range(self.num_experts):
            expert_layer = self.experts[expert_idx]
            idx, top_x = torch.where(expert_mask[expert_idx])

            # Index the correct hidden states and compute the expert hidden state for
            # the current expert. We need to make sure to multiply the output hidden
            # states by `routing_weights` on the corresponding tokens (top-1 and top-2)
            current_state = hidden_states[None, top_x].reshape(-1, self.in_features)
            if current_state.numel() == 0:
                continue
            current_hidden_states = (
                expert_layer(current_state) * routing_weights[top_x, idx, None]
            )

            # However `index_add_` only support torch tensors for indexing so we'll use
            # the `top_x` tensor here.
            final_hidden_states.index_add_(
                0, top_x, current_hidden_states.to(hidden_states.dtype)
            )
        final_hidden_states = final_hidden_states.reshape(
            *input_shape[:-1], self.out_features
        )
        final_hidden_states = pretrained_out + final_hidden_states
        return final_hidden_states

    @property
    def weight(self):
        """
        Mimic linear layer. Bacause in some cases, user might indicate the device (or dtype of parameters) of the linear layer using `linear_layer.weight.device`
        """
        return self.pretrained_model.weight

    @property
    def bias(self):
        return self.pretrained_model.bias

    def __repr__(self):
        return (
            f"SingularMoELinear("
            f"in_features={self.pretrained_model.in_features}, "
            f"out_features={self.pretrained_model.out_features}, "
            f"num_experts={self.num_experts}, "
            f"top_k={self.top_k}, "
            f"gate_k={self.gate_k}, "
            f"k={self.k}"
            f")"
        )



def _is_list_of_dict(obj) -> bool:
    if not isinstance(obj, list):
        return False
    for i in obj:
        if not isinstance(i, dict):
            return False
    return True


def _sprint_json_entry(obj):
    if isinstance(obj, str):
        return "string"
    elif isinstance(obj, float):
        return "float"
    elif isinstance(obj, int):
        return "int"
    elif isinstance(obj, list):
        if len(obj) > 0:
            return f"list[{_sprint_json_entry(obj[0])}]"
        else:
            return "list"
    else:
        return type(obj)


def print_json(j: dict, indent="  ", verbose: bool = False, print_type: bool = True):
    R"""print an overview of json file

    Examples:
        >>> print_json(open('path_to_json', 'r'))

    Args:
        j (dict): loaded json file
        indent (str, optional): Defaults to '  '.
    """

    def _print_json(j: dict, level):
        def _sprint(s):
            return indent * level + s

        for k in j.keys():
            if isinstance(j[k], dict):
                print(_sprint(k) + ":")
                _print_json(j[k], level + 1)
            elif _is_list_of_dict(j[k]):
                if verbose:
                    print(_sprint(k) + ": [")
                    for i in range(len(j[k]) - 1):
                        _print_json(j[k][0], level + 2)
                        print(_sprint(f"{indent},"))
                    _print_json(j[k][-1], level + 2)
                    print(_sprint(f"{indent}]"))
                else:
                    print(_sprint(k) + ": [")
                    _print_json(j[k][0], level + 2)
                    print(_sprint(f"{indent}] ... {len(j[k]) - 1} more"))
            else:
                if print_type:
                    print(f"{_sprint(k)}: {_sprint_json_entry(j[k])}")
                else:
                    print(f"{_sprint(k)}: {j[k]}")

    _print_json(j, level=0)



# flake8: noqa F401
from .clip_concrete_adamerging import (
    ConcreteLayerWiseAdaMergingForCLIP,
    ConcreteTaskWiseAdaMergingForCLIP,
)
from .clip_concrete_task_arithmetic import ConcreteTaskArithmeticAlgorithmForCLIP



import logging
from copy import deepcopy
from typing import Dict, List, Mapping, Optional, Union

import torch
from torch import nn

from fusion_bench.method.base_algorithm import BaseAlgorithm
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.state_dict_arithmetic import (
    state_dict_add,
    state_dict_avg,
    state_dict_div,
    state_dict_mul,
)
from fusion_bench.utils.type import StateDictType

log = logging.getLogger(__name__)


def simple_average(
    modules: List[Union[nn.Module, StateDictType]],
    base_module: Optional[nn.Module] = None,
):
    R"""
    Averages the parameters of a list of PyTorch modules or state dictionaries.

    This function takes a list of PyTorch modules or state dictionaries and returns a new module with the averaged parameters, or a new state dictionary with the averaged parameters.

    Args:
        modules (List[Union[nn.Module, StateDictType]]): A list of PyTorch modules or state dictionaries.
        base_module (Optional[nn.Module]): A base module to use for the new module. If provided, the averaged parameters will be loaded into this module. If not provided, a new module will be created by copying the first module in the list.

    Returns:
        module_or_state_dict (Union[nn.Module, StateDictType]): A new PyTorch module with the averaged parameters, or a new state dictionary with the averaged parameters.

    Examples:
        >>> import torch.nn as nn
        >>> model1 = nn.Linear(10, 10)
        >>> model2 = nn.Linear(10, 10)
        >>> averaged_model = simple_averageing([model1, model2])

        >>> state_dict1 = model1.state_dict()
        >>> state_dict2 = model2.state_dict()
        >>> averaged_state_dict = simple_averageing([state_dict1, state_dict2])
    """
    if isinstance(modules[0], nn.Module):
        if base_module is None:
            new_module = deepcopy(modules[0])
        else:
            new_module = base_module
        state_dict = state_dict_avg([module.state_dict() for module in modules])
        new_module.load_state_dict(state_dict)
        return new_module
    elif isinstance(modules[0], Mapping):
        return state_dict_avg(modules)


class SimpleAverageAlgorithm(
    BaseAlgorithm,
    SimpleProfilerMixin,
):
    @torch.no_grad()
    def run(self, modelpool: Union[BaseModelPool, Dict[str, nn.Module]]):
        """
        Fuse the models in the given model pool using simple averaging.

        This method iterates over the names of the models in the model pool, loads each model, and appends it to a list.
        It then returns the simple average of the models in the list.

        Args:
            modelpool: The pool of models to fuse.

        Returns:
            The fused model obtained by simple averaging.
        """
        if isinstance(modelpool, dict):
            modelpool = BaseModelPool(modelpool)

        log.info(
            f"Fusing models using simple average on {len(modelpool.model_names)} models."
            f"models: {modelpool.model_names}"
        )
        sd: Optional[StateDictType] = None
        forward_model = None
        merged_model_names = []

        for model_name in modelpool.model_names:
            with self.profile("load model"):
                model = modelpool.load_model(model_name)
                merged_model_names.append(model_name)
                print(f"load model of type: {type(model).__name__}")
            with self.profile("merge weights"):
                if sd is None:
                    # Initialize the state dictionary with the first model's state dictionary
                    sd = model.state_dict(keep_vars=True)
                    forward_model = model
                else:
                    # Add the current model's state dictionary to the accumulated state dictionary
                    sd = state_dict_add(sd, model.state_dict(keep_vars=True))
        with self.profile("merge weights"):
            # Divide the accumulated state dictionary by the number of models to get the average
            sd = state_dict_div(sd, len(modelpool.model_names))

        forward_model.load_state_dict(sd)
        # print profile report and log the merged models
        self.print_profile_summary()
        log.info(f"merged {len(merged_model_names)} models:")
        for model_name in merged_model_names:
            log.info(f"  - {model_name}")
        return forward_model



import logging
import os
from typing import Dict, List, Optional, cast

import numpy as np
import pandas as pd
import torch
import torch.utils
from numpy.typing import NDArray
from torch import nn
from tqdm.auto import tqdm

from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.parameters import (
    StateDictType,
    state_dict_to_vector,
    trainable_state_dict,
)
from fusion_bench.utils.state_dict_arithmetic import state_dict_sub

log = logging.getLogger(__name__)


class TaskVectorCosSimilarity(BaseAlgorithm, LightningFabricMixin):
    """
    This class is similar to the Dummy algorithm,
    but it also print (or save) the cosine similarity matrix between the task vectors of the models in the model pool.
    """

    _config_mapping = BaseAlgorithm._config_mapping | {
        "plot_heatmap": "plot_heatmap",
        "_output_path": "output_path",
    }

    def __init__(
        self,
        plot_heatmap: bool,
        trainable_only: bool = True,
        max_points_per_model: Optional[int] = None,
        output_path: Optional[str] = None,
        **kwargs,
    ):
        self.plot_heatmap = plot_heatmap
        self.trainable_only = trainable_only
        self.max_points_per_model = max_points_per_model
        self._output_path = output_path
        super().__init__(**kwargs)

    @property
    def output_path(self):
        if self._output_path is None:
            return self.fabric.logger.log_dir
        else:
            return self._output_path

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        pretrained_model = modelpool.load_pretrained_model()

        task_vectors = []
        for name, finetuned_model in tqdm(
            modelpool.named_models(), total=len(modelpool)
        ):
            print(f"computing task vectors for {name}")
            task_vectors.append(
                self.get_task_vector(pretrained_model, finetuned_model).to(
                    torch.float64
                )
            )
        task_vectors = torch.stack(task_vectors, dim=0)

        cos_sim_matrix = torch.zeros(
            len(modelpool), len(modelpool), dtype=torch.float64
        )
        for i in range(len(modelpool)):
            for j in range(i, len(modelpool)):
                assert task_vectors[i].size() == task_vectors[j].size()
                cos_sim_matrix[i, j] = torch.nn.functional.cosine_similarity(
                    task_vectors[i], task_vectors[j], dim=0
                )
                cos_sim_matrix[j, i] = cos_sim_matrix[i, j]

        # convert the matrix to a pandas DataFrame
        cos_sim_df = pd.DataFrame(
            cos_sim_matrix.numpy(),
            index=modelpool.model_names,
            columns=modelpool.model_names,
        )

        print(cos_sim_df)
        if self.output_path is not None:
            os.makedirs(self.output_path, exist_ok=True)
            cos_sim_df.to_csv(
                os.path.join(self.output_path, "task_vector_cos_similarity.csv")
            )

        if self.plot_heatmap:
            self._plot_heatmap(cos_sim_df)

        return pretrained_model

    def _plot_heatmap(self, data: pd.DataFrame):
        """
        This function plots a heatmap of the provided data using seaborn.

        Args:
            data (pd.DataFrame): A pandas DataFrame containing the data to be plotted.
            figsize (tuple): A tuple specifying the size of the figure. Default is (4, 3).

        Returns:
            None
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        # Create a heatmap using seaborn
        plt.figure()
        sns.heatmap(
            data,
            annot=True,
            fmt=".2f",
            cmap="GnBu",
        )

        # Add title and labels with increased font size
        plt.title("Heatmap of Cos Similarities", fontsize=14)
        # plt.xlabel("Task", fontsize=14)
        # plt.ylabel("Task", fontsize=14)
        plt.xticks(rotation=45)
        plt.yticks(rotation=45)

        # Show plot
        plt.savefig(
            os.path.join(self.output_path, "task_vector_cos_similarity.pdf"),
            bbox_inches="tight",
        )
        plt.close()

    def get_task_vector(
        self, pretrained_model: nn.Module, finetuned_model: nn.Module
    ) -> torch.Tensor:
        task_vector = state_dict_sub(
            self.get_state_dict(finetuned_model),
            self.get_state_dict(pretrained_model),
        )
        task_vector = state_dict_to_vector(task_vector)

        task_vector = task_vector.cpu().float().numpy()
        # downsample if necessary
        if (
            self.max_points_per_model is not None
            and self.max_points_per_model > 0
            and task_vector.shape[0] > self.max_points_per_model
        ):
            log.info(
                f"Downsampling task vectors to {self.max_points_per_model} points."
            )
            indices = np.random.choice(
                task_vector.shape[0], self.max_points_per_model, replace=False
            )
            task_vector = task_vector[indices].copy()

        task_vector = torch.from_numpy(task_vector)
        return task_vector

    def get_state_dict(self, model: nn.Module):
        if self.trainable_only:
            return trainable_state_dict(model)
        else:
            return model.state_dict()



import logging
from copy import deepcopy
from typing import Optional

from omegaconf import DictConfig, open_dict
from transformers import CLIPModel, CLIPProcessor, CLIPVisionModel
from typing_extensions import override

from fusion_bench.utils import instantiate, timeit_context

from ..base_pool import BaseModelPool

log = logging.getLogger(__name__)


class CLIPVisionModelPool(BaseModelPool):
    """
    A model pool for managing Hugging Face's CLIP Vision models.

    This class extends the base `ModelPool` class and overrides its methods to handle
    the specifics of the CLIP Vision models provided by the Hugging Face Transformers library.
    """

    _config_mapping = BaseModelPool._config_mapping | {"_processor": "processor"}

    def __init__(
        self,
        models: DictConfig,
        *,
        processor: Optional[DictConfig] = None,
        **kwargs,
    ):
        super().__init__(models, **kwargs)

        self._processor = processor

    def load_processor(self, *args, **kwargs) -> CLIPProcessor:
        assert self._processor is not None, "Processor is not defined in the config"
        processor = instantiate(self._processor, *args, **kwargs)
        return processor

    def load_clip_model(self, model_name: str, *args, **kwargs) -> CLIPModel:
        model_config = self._models[model_name]
        assert isinstance(model_config, DictConfig), "Model config must be a DictConfig"
        model_config = deepcopy(model_config)
        with open_dict(model_config):
            model_config._target_ = "transformers.CLIPModel.from_pretrained"
        clip_model = instantiate(model_config, *args, **kwargs)
        return clip_model

    @override
    def save_model(self, model: CLIPVisionModel, path: str):
        """
        Save a CLIP Vision model to the given path.

        Args:
            model (CLIPVisionModel): The model to save.
            path (str): The path to save the model to.
        """
        with timeit_context(f'Saving clip vision model to "{path}"'):
            model.save_pretrained(path)



import torch
import torch.nn as nn
import torch.nn.functional as F


class DeepLabHead(nn.Sequential):
    def __init__(self, in_channels: int, num_classes: int):
        super(DeepLabHead, self).__init__(
            ASPP(in_channels, [12, 24, 36]),
            nn.Conv2d(256, 256, 3, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.Conv2d(256, num_classes, 1),
        )


class ASPPConv(nn.Sequential):
    def __init__(self, in_channels, out_channels, dilation):
        modules = [
            nn.Conv2d(
                in_channels,
                out_channels,
                3,
                padding=dilation,
                dilation=dilation,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
        ]
        super(ASPPConv, self).__init__(*modules)


class ASPPPooling(nn.Sequential):
    def __init__(self, in_channels, out_channels):
        super(ASPPPooling, self).__init__(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
        )

    def forward(self, x):
        size = x.shape[-2:]
        x = super(ASPPPooling, self).forward(x)
        return F.interpolate(x, size=size, mode="bilinear", align_corners=False)


class ASPP(nn.Module):
    def __init__(self, in_channels: int, atrous_rates):
        super(ASPP, self).__init__()
        out_channels = 256
        modules = []
        modules.append(
            nn.Sequential(
                nn.Conv2d(in_channels, out_channels, 1, bias=False),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(),
            )
        )

        rate1, rate2, rate3 = tuple(atrous_rates)
        modules.append(ASPPConv(in_channels, out_channels, rate1))
        modules.append(ASPPConv(in_channels, out_channels, rate2))
        modules.append(ASPPConv(in_channels, out_channels, rate3))
        modules.append(ASPPPooling(in_channels, out_channels))

        self.convs = nn.ModuleList(modules)

        self.project = nn.Sequential(
            nn.Conv2d(5 * out_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
            nn.Dropout(0.5),
        )

    def forward(self, x):
        res = []
        for conv in self.convs:
            res.append(conv(x))
        res = torch.cat(res, dim=1)
        return self.project(res)



from transformers import AutoConfig, AutoModel, AutoModelForCausalLM

from .configuration_smile_mistral import SmileMistralConfig
from .modeling_smile_mistral import SmileMistralForCausalLM, SmileMistralModel

AutoConfig.register("smile_mistral", SmileMistralConfig)
AutoModel.register(SmileMistralConfig, SmileMistralModel)
AutoModelForCausalLM.register(SmileMistralConfig, SmileMistralForCausalLM)



from typing import Dict, Literal, Optional, Union  # noqa: F401

import torch
from torch import nn
from tqdm.auto import tqdm
from transformers import LlamaForCausalLM, LlamaModel

from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import CausalLMPool

from . import prune_utils
from .prune_utils import PruningType, find_linear_layers


def unstructured_magnitude_prune_(
    model: Union[LlamaForCausalLM, LlamaModel], sparsity_ratio: float
):
    """
    Perform unstructured magnitude pruning on the given model.

    Args:
        model (Union[LlamaForCausalLM, LlamaModel]): The model to be pruned.
        sparsity_ratio (float): The ratio of weights to be pruned.

    Returns:
        The pruned model.
    """
    if isinstance(model, LlamaForCausalLM):
        layers = model.model.layers
    elif isinstance(model, LlamaModel):
        layers = model.layers

    subset: Dict[str, nn.Linear] = find_linear_layers(layers)
    for name in tqdm(subset, desc="Pruning"):
        prune_utils.unstructured_magnitude_prune_(
            subset[name].weight,
            metric_function_or_scores=torch.rand_like,
            sparsity_ratio=sparsity_ratio,
        )

    return model


def semistructured_magnitude_prune_(
    model: Union[LlamaForCausalLM, LlamaModel], n: int, m: int
):
    """
    Perform semi-structured (N:M structured) magnitude pruning on the given model.

    Args:
        model (Union[LlamaForCausalLM, LlamaModel]): The model to be pruned.
        n (int): The number of weights to be pruned in each group.
        m (int): The total number of weights in each group.

    Returns:
        The pruned model.
    """
    if isinstance(model, LlamaForCausalLM):
        layers = model.model.layers
    elif isinstance(model, LlamaModel):
        layers = model.layers

    subset: Dict[str, nn.Linear] = find_linear_layers(layers)
    for name in tqdm(subset, desc="Pruning"):
        prune_utils.semistructured_magnitude_prune_(
            subset[name].weight,
            metric_function_or_scores=torch.rand_like,
            n=n,
            m=m,
        )

    return model


class RandomPruningForLlama(BaseAlgorithm, SimpleProfilerMixin):
    """
    A class to perform random pruning for Llama models.

    Attributes:
        prune_type (PruningType): The type of pruning to be performed.
        sparsity_ratio (float): The ratio of weights to be pruned.
        n (int): The number of weights to be pruned in each group (for semistructured pruning).
        m (int): The total number of weights in each group (for semistructured pruning).
    """

    _config_mapping = BaseAlgorithm._config_mapping | {
        "prune_type": "prune_type",
        "sparsity_ratio": "sparsity_ratio",
        "n": "n",
        "m": "m",
    }

    def __init__(
        self,
        *,
        prune_type: PruningType,
        sparsity_ratio: float,
        n: int,
        m: int,
        **kwargs,
    ):
        """
        Initialize the RandomPruningForLlama class.

        Args:
            prune_type (PruningType): The type of pruning to be performed.
            sparsity_ratio (float): The ratio of weights to be pruned.
            n (int): The number of weights to be pruned in each group (for semistructured pruning).
            m (int): The total number of weights in each group (for semistructured pruning).
            **kwargs: Additional keyword arguments.
        """
        self.prune_type = prune_type
        self.sparsity_ratio = sparsity_ratio
        self.n = n
        self.m = m
        super().__init__(**kwargs)

    @torch.no_grad()
    def run(self, modelpool: CausalLMPool):
        """
        Run the pruning algorithm on the first model from the given model pool.

        Args:
            modelpool (CausalLMPool): The pool of models to be pruned.

        Returns:
            The pruned model.
        """
        # load pre-trained model or the first model in the pool
        base_model = modelpool.load_pretrained_or_first_model()

        if self.prune_type == PruningType.UNSTRUCTURED:
            unstructured_magnitude_prune_(base_model, self.sparsity_ratio)
        elif self.prune_type == PruningType.SEMISTRUCTURED:
            semistructured_magnitude_prune_(base_model, self.n, self.m)
        else:
            raise ValueError(
                f"Invalid pruning type: {self.prune_type}"
                "Choose from 'unstructured' or 'semistructured'"
            )

        return base_model



import logging
import os
import re
from copy import deepcopy
from typing import Dict, List, Tuple

import torch
import torch.nn.functional as F
from accelerate import init_empty_weights
from torch import Tensor, nn
from tqdm.auto import tqdm
from transformers import AutoConfig, AutoTokenizer, MistralForCausalLM
from transformers.models.mistral.modeling_mistral import MistralDecoderLayer

from fusion_bench.compat.method import ModelFusionAlgorithm
from fusion_bench.compat.modelpool import to_modelpool
from fusion_bench.method.simple_average import simple_average
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.models.modeling_smile_mistral import (
    SmileMistralConfig,
    SmileMistralForCausalLM,
)
from fusion_bench.models.modeling_smile_mistral.modeling_smile_mistral import (
    SmileLinear,
    SmileMistralDecoderLayer,
)
from fusion_bench.models.utils import get_attr, set_attr
from fusion_bench.utils.dtype import parse_dtype
from fusion_bench.utils.parameters import print_parameters
from fusion_bench.utils.state_dict_arithmetic import state_dict_sub

log = logging.getLogger(__name__)


class ExpertNotTrainedError(Exception):
    pass


def _is_all_zeros(tensor: Tensor | List[Tensor]) -> bool:
    """
    Check if a tensor or a list of tensors are all zeros.

    Args:
        tensor (Tensor | List[Tensor]): The tensor or list of tensors to check.

    Returns:
        bool: True if all elements are zeros, False otherwise.
    """
    if isinstance(tensor, Tensor):
        return torch.allclose(tensor, torch.zeros_like(tensor))
    else:
        return all(_is_all_zeros(t) for t in tensor)


def _svd(w: Tensor, full_matrices=False) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Perform Singular Value Decomposition (SVD) on a tensor.

    Args:
        w (Tensor): The input tensor.
        full_matrices (bool, optional): Whether to compute the full-sized U and V matrices. Defaults to False.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The U, S, and V matrices from SVD.
    """
    device = w.device
    if w.device != torch.float32 or w.device != torch.float64:
        w = w.float()

    u, s, vh = torch.linalg.svd(
        w,
        full_matrices=full_matrices,
        # driver="gesvd" if w.is_cuda else None
    )
    v = vh.T

    u = u.to(device)
    s = s.to(device)
    v = v.to(device)
    return u, s, v


def svd(
    w: Tensor, full_matrices=True, accelerator=None
) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Perform SVD on a tensor with optional acceleration.

    Args:
        w (Tensor): The input tensor.
        full_matrices (bool, optional): Whether to compute the full-sized U and V matrices. Defaults to True.
        accelerator (optional): The device to perform the computation on. Defaults to None.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The U, S, and V matrices from SVD.
    """
    if accelerator is None:
        return _svd(w, full_matrices=full_matrices)
    original_device = w.device
    w = w.to(accelerator)
    u, s, v = _svd(w)
    return u, s, v


@torch.no_grad()
def upscale_to_smile_linear(
    base: nn.Linear, experts: List[nn.Linear], target: SmileLinear, accelerator=None
):
    """
    Upscale a base linear layer to a SmileLinear layer using expert models.

    Args:
        base (nn.Linear): The base linear layer.
        experts (List[nn.Linear]): A list of expert linear layers.
        target (SmileLinear): The target SmileLinear layer.
        accelerator (optional): The device to perform the computation on. Defaults to None.

    Returns:
        SmileLinear: The upscaled SmileLinear layer.
    """
    w = base.weight
    w_ft_list = [e.weight for e in experts]
    dw_list = [w_ft - w for w_ft in w_ft_list]

    if _is_all_zeros(dw_list):
        raise ExpertNotTrainedError("Expert models are not trained")

    rank_of_router = target.rank_of_router
    rank_of_expert = target.rank_of_expert
    num_local_experts = target.num_local_experts
    svd_list = [svd(dw, accelerator=accelerator) for dw in dw_list]

    # gate
    gate_weight = []
    for u, s, v in svd_list:
        gate_weight.append(v[:, :rank_of_router].T)
    gate_weight = (
        torch.stack(gate_weight, dim=0)
        .reshape(num_local_experts * rank_of_router, -1)
        .contiguous()
    )

    target.gate.load_state_dict({"weight": gate_weight})

    # shared linear
    target.shared_linear.load_state_dict(base.state_dict())

    # experts
    if rank_of_expert > 0:
        for expert_idx, target_expert in enumerate(target.experts):
            u, s, v = svd_list[expert_idx]
            u = u[:, :rank_of_expert]
            s = s[:rank_of_expert]
            v = v[:, :rank_of_expert]
            state_dict = {"u": u, "svh": (s * v).T}
            if experts[expert_idx].bias is not None:
                state_dict["bias"] = experts[expert_idx].bias.data
            target_expert.load_state_dict(state_dict)
    else:
        for expert_idx, target_expert in enumerate(target.experts):
            target_expert.load_state_dict(
                state_dict_sub(experts[expert_idx].state_dict(), base.state_dict())
            )

    return target


class SmileMistralUpscalingAlgorithm(ModelFusionAlgorithm, SimpleProfilerMixin):
    R"""
    SmileMistralUpscalingAlgorithm is a model fusion algorithm designed to upscale
    a pretrained Mistral model using a set of fine-tuned expert models. The algorithm
    leverages Singular Value Decomposition (SVD) to merge the weights of the pretrained
    model and the expert models into a new upscaled model.

    Attributes:
        modelpool (BaseModelPool): The pool of models to be used for upscaling.
        config (dict): Configuration parameters for the upscaling process.

    Methods:
        run(modelpool: BaseModelPool) -> SmileMistralForCausalLM:
            Executes the upscaling process and returns the upscaled model.

        merge(pretrained_model: MistralForCausalLM, finetuned_models: List[MistralForCausalLM]) -> SmileMistralForCausalLM:
            Merges the pretrained model with the fine-tuned models to create an upscaled model.
    """

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool) -> SmileMistralForCausalLM:
        """
        Executes the upscaling process.

        Args:
            modelpool (ModelPool): The pool of models to be used for upscaling.

        Returns:
            SmileMistralForCausalLM: The upscaled model.
        """
        self.modelpool = modelpool = to_modelpool(modelpool)
        config = self.config

        print(config)
        if config.model_path is not None and os.path.exists(config.model_path):
            log.info(f"Loading model from {config.model_path}")
            model = torch.load(config.model_path)
            print_parameters(model)
            return model

        with self.profile("load pretrained model"):
            pretrained_model = modelpool.load_model("_pretrained_")
        with self.profile("load fine-tuned model"):
            finetuned_models = [
                m for m in tqdm(modelpool.models(), total=len(modelpool.model_names))
            ]

        if config.device == "cuda" and torch.cuda.is_available():
            pretrained_model = pretrained_model.cuda()
            finetuned_models = [m.cuda() for m in finetuned_models]

        with self.profile("merge model"):
            model = self.merge(pretrained_model, finetuned_models)

        self.print_profile_summary()
        print_parameters(model)
        print(model)

        if config.model_dtype is not None:
            model.to(dtype=parse_dtype(config.model_dtype))

        if config.model_path is not None:
            if os.path.dirname(config.model_path):
                os.makedirs(os.path.dirname(config.model_path), exist_ok=True)
            log.info(f"Saving model to {config.model_path}")
            pretrained_model_config = self.modelpool.get_model_config("_pretrained_")
            pretrained_path = pretrained_model_config.get(
                "path", pretrained_model_config["pretrained_model_name_or_path"]
            )
            tokenizer = AutoTokenizer.from_pretrained(pretrained_path)
            tokenizer.save_pretrained(config.model_path)
            model.save_pretrained(config.model_path)

        return model

    def merge(
        self,
        pretrained_model: MistralForCausalLM,
        finetuned_models: List[MistralForCausalLM],
    ):
        """
        Merges the pretrained model with the fine-tuned models to create an upscaled model.

        Args:
            pretrained_model (MistralForCausalLM): The pretrained model.
            finetuned_models (List[MistralForCausalLM]): A list of fine-tuned models.

        Returns:
            SmileMistralForCausalLM: The upscaled model.
        """
        config = self.config

        with init_empty_weights():
            pretrained_model_config = self.modelpool.get_model_config("_pretrained_")
            pretrained_path = pretrained_model_config.get(
                "path", pretrained_model_config["pretrained_model_name_or_path"]
            )
            base_config = AutoConfig.from_pretrained(pretrained_path)
            model_config = SmileMistralConfig(
                num_experts_per_tok=config.num_experts_per_tok,
                rank_of_router=config.rank_of_router,
                rank_of_expert=config.rank_of_expert,
                num_local_experts=len(finetuned_models),
                **base_config.to_dict(),
            )
            model = SmileMistralForCausalLM(model_config)

        model.to(dtype=pretrained_model.dtype).to_empty(device="cpu")

        # copy pretrained model weights
        state_dict = model.state_dict()
        pretrained_state_dict = dict(pretrained_model.state_dict())
        for key in list(pretrained_state_dict.keys()):
            if key not in state_dict:
                pretrained_state_dict.pop(key)
        model.load_state_dict(pretrained_state_dict, strict=False)

        # upscale model
        for layer_idx in tqdm(
            range(len(pretrained_model.model.layers)),
            "Upscaling Modules (layer)",
            dynamic_ncols=True,
        ):
            pretrained_layer: MistralDecoderLayer = pretrained_model.model.layers[
                layer_idx
            ]
            finetuned_layers: List[MistralDecoderLayer] = [
                m.model.layers[layer_idx] for m in finetuned_models
            ]

            target_layer: SmileMistralDecoderLayer = model.model.layers[layer_idx]

            for n in ["q_proj", "k_proj", "v_proj", "o_proj"]:
                try:
                    upscale_to_smile_linear(
                        base=getattr(pretrained_layer.self_attn, n),
                        experts=[getattr(m.self_attn, n) for m in finetuned_layers],
                        target=getattr(target_layer.self_attn, n),
                        accelerator=config.accelerator,
                    )
                except ExpertNotTrainedError:
                    setattr(
                        target_layer.self_attn,
                        n,
                        getattr(pretrained_layer.self_attn, n),
                    )

            for n in ["gate_proj", "up_proj", "down_proj"]:
                try:
                    upscale_to_smile_linear(
                        base=getattr(pretrained_layer.mlp, n),
                        experts=[getattr(m.mlp, n) for m in finetuned_layers],
                        target=getattr(target_layer.mlp, n),
                        accelerator=config.accelerator,
                    )
                except ExpertNotTrainedError:
                    setattr(
                        target_layer.mlp,
                        n,
                        getattr(pretrained_layer.mlp, n),
                    )

        return model



from contextlib import contextmanager
from typing import Generator

from lightning.fabric.utilities.rank_zero import rank_zero_only
from lightning.pytorch.profilers import SimpleProfiler

__all__ = ["SimpleProfilerMixin"]


class SimpleProfilerMixin:
    """
    A mixin class that provides simple profiling capabilities.

    This mixin allows for easy profiling of code blocks using a context manager.
    It also provides methods to start and stop profiling actions, and to print
    a summary of the profiling results.

    Examples:

    ```python
    class MyClass(SimpleProfilerMixin):
        def do_something(self):
            with self.profile("work"):
                # do some work here
                ...
            with self.profile("more work"):
                # do more work here
                ...

            # print the profiling summary
            self.print_profile_summary()
    ```

    Attributes:
        _profiler (SimpleProfiler): An instance of the SimpleProfiler class used for profiling.
    """

    _profiler: SimpleProfiler = None

    @property
    def profiler(self):
        # Lazy initialization of the profiler instance
        if self._profiler is None:
            self._profiler = SimpleProfiler()
        return self._profiler

    @contextmanager
    def profile(self, action_name: str) -> Generator:
        """
        Context manager for profiling a code block

        Example:

        ```python
        with self.profile("work"):
            # do some work here
            ...
        ```
        """
        try:
            self.start_profile(action_name)
            yield action_name
        finally:
            self.stop_profile(action_name)

    def start_profile(self, action_name: str):
        self.profiler.start(action_name)

    def stop_profile(self, action_name: str):
        self.profiler.stop(action_name)

    @rank_zero_only
    def print_profile_summary(self):
        print(self.profiler.summary())

    def __del__(self):
        if self._profiler is not None:
            del self._profiler
            self._profiler = None



from typing import Literal

import torch.nn as nn

from . import resnet


class ResnetDilated(nn.Module):
    def __init__(self, orig_resnet, dilate_scale=8):
        super(ResnetDilated, self).__init__()
        from functools import partial

        if dilate_scale == 8:
            orig_resnet.layer3.apply(partial(self._nostride_dilate, dilate=2))
            orig_resnet.layer4.apply(partial(self._nostride_dilate, dilate=4))
        elif dilate_scale == 16:
            orig_resnet.layer4.apply(partial(self._nostride_dilate, dilate=2))

        # take pre-defined ResNet, except AvgPool and FC
        self.conv1 = orig_resnet.conv1
        self.bn1 = orig_resnet.bn1
        self.relu = orig_resnet.relu

        self.maxpool = orig_resnet.maxpool
        self.layer1 = orig_resnet.layer1
        self.layer2 = orig_resnet.layer2
        self.layer3 = orig_resnet.layer3
        self.layer4 = orig_resnet.layer4

        self.feature_dim = orig_resnet.feature_dim

    def _nostride_dilate(self, m, dilate):
        classname = m.__class__.__name__
        if classname.find("Conv") != -1:
            # the convolution with stride
            if m.stride == (2, 2):
                m.stride = (1, 1)
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate // 2, dilate // 2)
                    m.padding = (dilate // 2, dilate // 2)
            # other convoluions
            else:
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate, dilate)
                    m.padding = (dilate, dilate)

    def forward(self, x):
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        return x

    def forward_stage(self, x, stage):
        assert stage in [
            "conv",
            "layer1",
            "layer2",
            "layer3",
            "layer4",
            "layer1_without_conv",
        ]

        if stage == "conv":
            x = self.relu(self.bn1(self.conv1(x)))
            x = self.maxpool(x)
            return x

        elif stage == "layer1":
            x = self.relu(self.bn1(self.conv1(x)))
            x = self.maxpool(x)
            x = self.layer1(x)
            return x

        elif stage == "layer1_without_conv":
            x = self.layer1(x)
            return x

        else:  # Stage 2, 3 or 4
            layer = getattr(self, stage)
            return layer(x)


def resnet_dilated(
    basenet: str, pretrained: bool = True, dilate_scale: Literal[8, 16] = 8
):
    r"""Dilated Residual Network models from `"Dilated Residual Networks" <https://openaccess.thecvf.com/content_cvpr_2017/papers/Yu_Dilated_Residual_Networks_CVPR_2017_paper.pdf>`_

    Args:
        basenet (str): The type of ResNet.
        pretrained (bool): If True, returns a model pre-trained on ImageNet.
        dilate_scale ({8, 16}, default=8): The type of dilating process.
    """
    return ResnetDilated(
        resnet.__dict__[basenet](pretrained=pretrained), dilate_scale=dilate_scale
    )



from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import torch
from torch import Tensor
from torch.utils.hooks import RemovableHandle
from transformers import CLIPModel, CLIPProcessor, CLIPVisionModel
from transformers.models.clip.modeling_clip import CLIPVisionTransformer

from fusion_bench.models.hf_clip import HFCLIPClassifier
from fusion_bench.models.sparse_we_moe import (
    SparseWeightEnsemblingMoE,
    SparseWeightEnsemblingMoE_ShardGate,
)

from .taskpool import CLIPVisionModelTaskPool


class LayerWiseRoutingWeightSaver:
    def __init__(self, save_path: Path, max_num: Optional[int] = None):
        self.save_path = save_path
        self.max_num = max_num
        self.routing_weights = []

    def __call__(self, module, input: Tuple[Tensor], output: Tensor):
        assert isinstance(output, Tensor), "Output is expected to be a Tensor"
        # (batch_size, num_tokens, num_experts)
        routing_weights = output.detach().cpu()
        if self.max_num is not None and self.max_num > 0:
            if len(self.routing_weights) > self.max_num:
                return
            elif routing_weights.size(0) + len(self.routing_weights) > self.max_num:
                self.routing_weights.append(
                    routing_weights[: self.max_num - len(self.routing_weights)]
                )
            else:
                self.routing_weights.append(routing_weights)
        else:
            self.routing_weights.append(routing_weights)

    def save_routing_weights(self):
        routing_weights = torch.cat(self.routing_weights, dim=0)
        if self.save_path is not None:
            self.save_path.parent.mkdir(parents=True, exist_ok=True)
            print(f"Saving routing weights to {self.save_path}")
            torch.save(routing_weights, self.save_path)


class SparseWEMoECLIPVisionModelTaskPool(CLIPVisionModelTaskPool):

    # hooks and handles for saving layer-wise routing weights
    _layer_wise_routing_weights_save_hooks: Dict[Any, LayerWiseRoutingWeightSaver] = {}
    _layer_wise_routing_weights_save_hook_handles: Dict[Any, RemovableHandle] = {}

    _config_mapping = CLIPVisionModelTaskPool._config_mapping | {
        "_layer_wise_routing_weights_save_path": "layer_wise_routing_weights_save_path",
    }

    def __init__(
        self,
        layer_wise_routing_weights_save_path: Optional[str],
        layer_wise_routing_weights_max_num: Optional[int] = None,
        **kwargs,
    ):
        # save path for layer-wise routing weights
        self._layer_wise_routing_weights_save_path = (
            layer_wise_routing_weights_save_path
        )
        self.layer_wise_routing_weights_save_path = (
            Path(layer_wise_routing_weights_save_path)
            if layer_wise_routing_weights_save_path is not None
            else None
        )
        self.layer_wise_routing_weights_max_num = layer_wise_routing_weights_max_num
        super().__init__(**kwargs)

    def on_task_evaluation_begin(self, classifier: HFCLIPClassifier, task_name: str):
        super().on_task_evaluation_begin(classifier, task_name)
        if self.layer_wise_routing_weights_save_path is not None:
            # setup hooks for saving layer-wise routing weights
            assert isinstance(
                classifier.clip_model.vision_model,
                (CLIPVisionTransformer, CLIPVisionModel),
            ), "Vision model is expected to be a CLIPVisionTransformer"
            vision_model = classifier.clip_model.vision_model
            if isinstance(vision_model, CLIPVisionModel):
                vision_model = vision_model.vision_model
                # assign forward hooks for each layer
            shared_gate = None
            for i, layer in enumerate(vision_model.encoder.layers):
                mlp = layer.mlp
                assert isinstance(
                    mlp,
                    (SparseWeightEnsemblingMoE, SparseWeightEnsemblingMoE_ShardGate),
                ), f"MLP is expected to be a SparseWeightEnsemblingMoE or SparseWeightEnsemblingMoE_ShardGate, but got {type(mlp)}"
                # layer-wise routing weights
                hook = LayerWiseRoutingWeightSaver(
                    self.layer_wise_routing_weights_save_path
                    / task_name
                    / f"layer_{i}.pt",
                    max_num=self.layer_wise_routing_weights_max_num,
                )
                self._layer_wise_routing_weights_save_hooks[i] = hook
                if isinstance(mlp, SparseWeightEnsemblingMoE_ShardGate):
                    # if use shared gate, copy the gate to all layers to avoid multiple hooks
                    if shared_gate is None:
                        shared_gate = mlp.gate
                    mlp.gate = deepcopy(shared_gate)
                self._layer_wise_routing_weights_save_hook_handles[i] = (
                    mlp.gate.register_forward_hook(hook)
                )

    def on_task_evaluation_end(self):
        super().on_task_evaluation_end()
        if self.layer_wise_routing_weights_save_path is not None:
            # remove hooks for saving layer-wise routing weights
            for i, handle in self._layer_wise_routing_weights_save_hook_handles.items():
                self._layer_wise_routing_weights_save_hooks[i].save_routing_weights()
                handle.remove()



# flake8: noqa F401
from .llama_magnitude_prune import MagnitudePruningForLlama
from .llama_random_prune import RandomPruningForLlama
from .llama_wanda_prune import WandaPruningForLlama
from .magnitude_diff_pruning import MagnitudeDiffPruningAlgorithm



import logging

from omegaconf import DictConfig
from peft import PeftModel
from transformers import AutoModelForSeq2SeqLM

from fusion_bench.compat.modelpool.base_pool import ModelPool
from fusion_bench.utils import timeit_context

log = logging.getLogger(__name__)


class PeftModelForSeq2SeqLMPool(ModelPool):
    def load_model(self, model_config: str | DictConfig):
        """
        Load a model based on the provided configuration.

        The configuration options of `model_config` are:

        - name: The name of the model. If it is "_pretrained_", a pretrained Seq2Seq language model is returned.
        - path: The path where the model is stored.
        - is_trainable: A boolean indicating whether the model parameters should be trainable. Default is `True`.
        - merge_and_unload: A boolean indicating whether to merge and unload the PEFT model after loading. Default is `True`.


        Args:
            model_config (str | DictConfig): The configuration for the model. This can be either a string (name of the model) or a DictConfig object containing the model configuration.


        Returns:
            model: The loaded model. If the model name is "_pretrained_", it returns a pretrained Seq2Seq language model. Otherwise, it returns a PEFT model.
        """
        if isinstance(model_config, str):
            model_config = self.get_model_config(model_config)
        with timeit_context(f"Loading model {model_config['name']}"):
            if model_config["name"] == "_pretrained_":
                model = AutoModelForSeq2SeqLM.from_pretrained(model_config["path"])
                return model
            else:
                model = self.load_model("_pretrained_")
                peft_model = PeftModel.from_pretrained(
                    model,
                    model_config["path"],
                    is_trainable=model_config.get("is_trainable", True),
                )
                if model_config.get("merge_and_unload", True):
                    return peft_model.merge_and_unload()
                else:
                    return peft_model



import itertools
import logging
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

import torch

from fusion_bench import BaseTaskPool
from fusion_bench.taskpool.dummy import get_model_summary
from fusion_bench.utils.devices import get_device
from fusion_bench.utils.rich_utils import print_bordered

if TYPE_CHECKING:
    from transformers import LlamaForCausalLM, PreTrainedTokenizer

    from fusion_bench.modelpool import CausalLMPool
log = logging.getLogger(__name__)


def generate_text(
    model: "LlamaForCausalLM",
    tokenizer: "PreTrainedTokenizer",
    prompt: str,
    max_length: int = 1024,
    temperature: float = 0.01,
    top_p=0.9,
    device: torch.device = None,
):
    """
    Generate text using the loaded model.

    Args:
        model: The loaded language model
        tokenizer: The loaded tokenizer
        prompt (str): Input prompt text
        max_length (int): Maximum length of generated sequence
        temperature (float): Controls randomness (higher = more random)
        top_p (float): Nucleus sampling parameter

    Returns:
        str: Generated text
    """
    if device is None:
        device = get_device(model)

    # Encode the prompt
    inputs = tokenizer(prompt, return_tensors="pt")

    # Move to GPU if available
    inputs = {k: v.to(device) for k, v in inputs.items()}

    # Generate
    with torch.no_grad():
        outputs = model.generate(
            **inputs,
            max_length=max_length,
            temperature=temperature,
            top_p=top_p,
            pad_token_id=tokenizer.pad_token_id,
            eos_token_id=tokenizer.eos_token_id,
            do_sample=True,
        )

    # Decode and return the generated text
    generated_text = tokenizer.decode(outputs[0], skip_special_tokens=False)
    response = generated_text[len(prompt) :]
    return {
        "generated_text": generated_text,
        "response": response,
        "num_tokens": len(outputs[0]) - len(inputs["input_ids"][0]),
    }


class LlamaTestGenerationTaskPool(BaseTaskPool):
    """
    This task pool is used to evaluate a language model on a set of prompts.
    For the purpose of debugging, it can also be used in an interactive mode.
    """

    def __init__(
        self,
        test_prompts: List[str],
        max_length: int = 1024,
        temperature: float = 0.01,
        top_p: float = 0.9,
        iterative_mode: bool = False,
        **kwargs,
    ):
        """
        Args:
            test_prompts (List[str]): A list of prompts to be used for testing the model.
            max_length (int, optional): The maximum length of the generated text. Defaults to 1024.
            temperature (float, optional): The sampling temperature for text generation. Defaults to 0.01.
            top_p (float, optional): The cumulative probability for nucleus sampling. Defaults to 0.9.
            iterative_mode (bool, optional): If True, enables interactive mode for debugging. Defaults to False.
        """
        self.test_prompts = test_prompts
        self.max_length = max_length
        self.temperature = temperature
        self.top_p = top_p
        self.iterative_mode = iterative_mode
        super().__init__(**kwargs)

    def evaluate(
        self,
        model: Union["LlamaForCausalLM", Any],
        tokenizer: Optional["PreTrainedTokenizer"] = None,
    ):
        if tokenizer is None:
            if self._program is None:
                log.error(
                    "`_program` is not set. This is probably happening when you are not runing the program via `fusion_bench` CLI."
                    "Please pass `tokenizer` to this function."
                )
            modelpool: "CausalLMPool" = self._program.modelpool
            tokenizer = modelpool.load_tokenizer()

        report = get_model_summary(model)
        if self.test_prompts is not None:
            for prompt_idx, prompt in enumerate(self.test_prompts):
                print(f"=== Generating text {prompt_idx+1}/{len(self.test_prompts)}")
                report[f"conversation_{prompt_idx+1}"] = self._generate_text(
                    model, tokenizer, prompt
                )

        if self.iterative_mode:
            for prompt_idx in itertools.count():
                # Prompt for input
                # print usage instructions
                print("Enter a prompt to generate text. Type 'exit' to exit the loop.")
                prompt = input(
                    f"Enter a prompt, or type 'exit' to quit ({prompt_idx+1}): "
                )
                if prompt == "exit":
                    break
                report[f"iterative_conversation_{prompt_idx+1}"] = self._generate_text(
                    model, tokenizer, prompt
                )

        return report

    def _generate_text(
        self, model: "LlamaForCausalLM", tokenizer: "PreTrainedTokenizer", prompt: str
    ) -> dict:
        """
        Generate text using the provided model and tokenizer for a given prompt.

        This method generates text based on the given prompt using the specified model and tokenizer.
        It prints the prompt and the generated response, and returns a dictionary containing the prompt,
        response, wall time, number of characters, and number of tokens.

        Args:
            model: The language model to be used for text generation.
            tokenizer: The tokenizer to be used for encoding and decoding text.
            prompt (str): The input prompt for text generation.

        Returns:
            dict: A dictionary containing the following keys:
                - "prompt" (str): The input prompt.
                - "response" (str): The generated response.
                - "wall_time" (float): The time taken to generate the response.
                - "num_chars" (int): The number of characters in the generated response.
                - "num_tokens" (int): The number of tokens in the generated response.
        """
        print(prompt)
        start_time = time.time()
        outputs = generate_text(
            model,
            tokenizer=tokenizer,
            prompt=prompt,
            max_length=self.max_length,
            temperature=self.temperature,
            top_p=self.top_p,
        )
        print_bordered(
            outputs["response"], title="Generated Text", code_style="markdown"
        )
        print("\n")
        return {
            "prompt": prompt,
            "response": outputs["response"],
            "wall_time": time.time() - start_time,
            "num_chars": len(outputs["response"]),
            "num_tokens": outputs["num_tokens"],
        }



"""
Augmenters for ARC tasks

The augmenters are used to apply transformations to the input and output of the tasks.

Augmenter: Base class for all augmenters
Rotate: Rotate the grid by 90, 180, 270 degrees
PermuteColors: Permute the colors in the grid
PermuteColorswithMap: Permute the colors in the grid using a given color map
PermuteColorsRespectKeyColors: Permute the colors in the grid while keeping the key colors fixed
Flip: Flip the grid along the given axis
Reflect: Reflect the grid along the given axis
Repeat: Repeat the grid along the given
Transpose: Transpose the grid
IncreaseResolution: Increase the resolution of the grid
IncreaseHeight: Increase the height of the grid
IncreaseWidth: Increase the width of the grid
DropoutInput: Delete a random rectangular patch from the input
DropoutOutput: Delete a random rectangular patch from the output
RandomTranslateXY: Randomly shift the grid along the x and y axis
RandomTranslateX: Randomly shift the grid along the x axis
RandomTranslateY: Randomly shift the grid along the y axis
Chain: Chain multiple augmenters together
Concat: Concatenate the output of multiple augmenters
IdentityAugmenter: Identity augmenter
PermuteExamples: Permute the training examples in the task
"""

import copy
from typing import List, Optional, Tuple

import numpy as np
from numpy.random import RandomState

from .arc import Example, Grid, Task


class Augmenter:
    share_rng: bool = False

    def __repr__(self):
        return str(self)

    def __call__(self, grid: Grid, rng: RandomState = None) -> Grid:
        return self.apply_to_grid(grid, rng=rng)

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        raise NotImplementedError()

    def apply_to_example(
        self,
        example: Example,
        rng: RandomState = None,
        to_input: bool = True,
        to_output: bool = True,
    ) -> Example:
        input = (
            self.apply_to_grid(example.input, rng=rng) if to_input else example.input
        )
        output = (
            self.apply_to_grid(example.output, rng=rng) if to_output else example.output
        )
        if example.cot is not None:
            cot = copy.deepcopy(example.cot)
            if not np.array_equal(cot[-1], output):
                cot.append(output)
        else:
            cot = None
        return Example(input, output, cot)

    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:

        if self.share_rng:
            # make sure all examples get seperate copies of same rng
            # this is to make sure that the same random number(s) is used for all examples
            # seed = rng.randint(0, 2**32)
            train_rngs = [copy.deepcopy(rng) for i in range(len(task.train_examples))]
            test_rng = copy.deepcopy(rng)  # RandomState(seed)
        else:
            train_rngs = [rng for _ in range(len(task.train_examples))]
            test_rng = rng

        return Task(
            train_examples=[
                self.apply_to_example(example, rng=rng_i, **kwargs)
                for example, rng_i in zip(task.train_examples, train_rngs)
            ],
            test_example=self.apply_to_example(
                task.test_example, rng=test_rng, **kwargs
            ),
            name=task.name,
        )


class Rotate(Augmenter):
    def __init__(self, angle: int):
        assert angle in [0, 90, 180, 270]
        self.angle = angle

    def __str__(self):
        return f"Rotate({self.angle})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        # roate the input and output by the given angle
        del rng
        if self.angle == 90:
            return np.rot90(grid, k=1)
        elif self.angle == 180:
            return np.rot90(grid, k=2)
        elif self.angle == 270:
            return np.rot90(grid, k=3)
        else:
            raise ValueError("Invalid angle")


class PermuteColors(Augmenter):
    share_rng = True
    color_mapper = None

    def __str__(self):
        return "PermuteColors()"

    def apply_to_task(
        self,
        task: Task,
        use_test_output: bool = True,
        rng: RandomState = None,
        share_rng=False,
        **kwargs,
    ) -> Task:
        # get all used colors in the inputs and outputs
        colors = []
        for example in task.train_examples:
            colors += example.input.flatten().tolist()
            colors += example.output.flatten().tolist()
        colors += task.test_example.input.flatten().tolist()

        if use_test_output:
            colors += task.test_example.output.flatten().tolist()

        # get unique colors
        colors = set(colors)
        # remove 0
        colors = colors - {0}
        remaining_colors = list(set(list(range(1, 10))) - colors)
        colors = list(colors)

        rng.shuffle(remaining_colors)  # inplac

        permuted_colors = rng.permutation(colors).tolist()
        # sample a mapping from colors to new ids
        color_map = {0: 0}

        for color in colors:

            if color in color_map:
                continue

            if len(remaining_colors) > 0:
                new_color = remaining_colors.pop()
            else:
                new_color = permuted_colors.pop()
                # in this case we want to directly swap colors
                # so new_color shoud be the color that is being replaced
                color_map[new_color] = color
                # remove color from permuted colors
                if color in permuted_colors:
                    permuted_colors.remove(color)

            color_map[color] = new_color

        self._color_map = color_map

        def color_mapper(color: int) -> int:
            return color_map.get(color, color)

        self.color_mapper = np.vectorize(color_mapper)

        return Task(
            train_examples=[
                self.apply_to_example(example, rng=rng, **kwargs)
                for example in task.train_examples
            ],
            test_example=self.apply_to_example(task.test_example, rng=rng, **kwargs),
            name=task.name,
        )

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        # map colors to new colors
        return self.color_mapper(grid)


class PermuteColorswithMap(Augmenter):

    def __init__(self, color_map):
        self.color_map = color_map

        def color_mapper(color: int) -> int:
            return color_map.get(color, color)

        self.color_mapper = np.vectorize(color_mapper)

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        # map colors to new colors
        return self.color_mapper(grid)

    def __repr__(self):
        return f"PermuteColorswithMap({self.color_map})"

    def __str__(self):
        return self.__repr__()


class PermuteColorsRespectKeyColors(Augmenter):
    share_rng = True
    color_mapper = None

    def __init__(
        self,
        key_colors: List[int] = None,
        use_remaining_colors: bool = True,
        use_test_output: bool = False,
    ):
        self.key_colors = key_colors
        self.use_remaining_colors = use_remaining_colors
        self.use_test_output = use_test_output

    def __str__(self):
        return "PermuteColorsRespectKeyColors(key_colors={self.key_colors}, use_remaining_colors={self.use_remaining_colors}, use_test_output={self.use_test_output})"

    @staticmethod
    def get_key_colors(task: Task, use_test_output: bool = False):
        key_colors_input = []
        key_colors_output = []
        for example in task.train_examples:
            key_colors_input.append(set(example.input.flatten().tolist()))
            key_colors_output.append(set(example.output.flatten().tolist()))

        key_colors_input.append(set(task.test_example.input.flatten().tolist()))
        if use_test_output:
            key_colors_output.append(set(task.test_example.output.flatten().tolist()))

        all_colors = set.union(*key_colors_input, *key_colors_output)

        key_colors_input = set.intersection(*key_colors_input)
        key_colors_output = set.intersection(*key_colors_output)
        key_colors = key_colors_input.union(key_colors_output)

        return key_colors, all_colors

    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:
        key_colors, colors = PermuteColorsRespectKeyColors.get_key_colors(
            task, use_test_output=self.use_test_output
        )
        if self.key_colors is None:
            self.key_colors = key_colors
        else:
            key_colors = self.key_colors

        remaining_colors = list(set(list(range(10))) - colors)
        colors = list(colors)

        rng.shuffle(remaining_colors)  # inplace
        # keep key colors_static
        color_map = {}

        for key_color in self.key_colors:
            color_map[key_color] = key_color
            if key_color in colors:
                colors.remove(key_color)

        permuted_colors = rng.permutation(colors).tolist()

        for color in colors:

            if color in color_map:
                continue

            if self.use_remaining_colors and len(remaining_colors) > 0:
                new_color = remaining_colors.pop()
            else:
                new_color = permuted_colors.pop()
                # in this case we want to directly swap colors
                # so new_color shoud be the color that is being replaced
                color_map[new_color] = color
                # remove color from permuted colors
                if color in permuted_colors:
                    permuted_colors.remove(color)

            color_map[color] = new_color

        def color_mapper(color: int) -> int:
            return color_map.get(color, color)

        self.color_mapper = np.vectorize(color_mapper)

        return Task(
            train_examples=[
                self.apply_to_example(example, rng=rng, **kwargs)
                for example in task.train_examples
            ],
            test_example=self.apply_to_example(task.test_example, rng=rng, **kwargs),
            name=task.name,
        )

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        # map colors to new colors
        return self.color_mapper(grid)


class PermuteColorsStatic(Augmenter):
    share_rng = True
    color_mapper = None

    def __str__(self):
        return f"PermuteColorsStatic({self.color_map})"

    def __init__(self, color_map):
        self.color_map = color_map

        self.color_mapper = np.vectorize(lambda x: self.color_map.get(x, x))

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        return self.color_mapper(grid)


class Flip(Augmenter):
    def __init__(self, axis: int):
        assert axis in [0, 1]
        self.axis = axis

    def __str__(self):
        return f"Flip({self.axis})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        if self.axis == 0:
            return np.flipud(grid)
        elif self.axis == 1:
            return np.fliplr(grid)
        else:
            raise ValueError("Invalid axis")


class Reflect(Augmenter):
    def __init__(self, axis: int, reverse=False):
        assert axis in [0, 1]
        self.axis = axis
        self.reverse = reverse

    def __str__(self):
        return f"Reflect({self.axis}, reverse={self.reverse})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Example:
        del rng
        # reflect the input and output by the given axis
        original = grid
        if self.axis == 0:
            reflected = np.flipud(original)
        elif self.axis == 1:
            reflected = np.fliplr(original)
        else:
            raise ValueError("Invalid axis")

        if self.reverse:
            if self.axis == 0:
                return np.concatenate((reflected, original), axis=0)
            elif self.axis == 1:
                return np.concatenate((reflected, original), axis=1)
        else:
            if self.axis == 0:
                return np.concatenate((original, reflected), axis=0)
            elif self.axis == 1:
                return np.concatenate((original, reflected), axis=1)


class Repeat(Augmenter):
    def __init__(self, axis: int, n: int):
        assert axis in [0, 1, 2]
        self.axis = axis
        self.n = n

    def __str__(self):
        return f"Repeat({self.axis}, {self.n})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        if self.axis == 0:
            return np.concatenate([grid] * self.n, axis=0)
        elif self.axis == 1:
            return np.concatenate([grid] * self.n, axis=1)
        elif self.axis == 2:
            return np.concatenate(
                [np.concatenate([grid] * self.n, axis=0)] * self.n, axis=1
            )
        else:
            raise ValueError("Invalid axis")


class Transpose(Augmenter):
    def __str__(self):
        return "Transpose()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        return np.transpose(grid)


class IncreaseResolution(Augmenter):
    def __init__(self, factor: int):
        assert factor > 1
        self.factor = factor

    def __str__(self):
        return f"IncreaseResolution({self.factor})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        grid = np.repeat(grid, self.factor, axis=0)
        grid = np.repeat(grid, self.factor, axis=1)
        return grid


class IncreaseHeight(Augmenter):
    def __init__(self, factor: int):
        assert factor > 1
        self.factor = factor

    def __str__(self):
        return f"IncreaseHeight({self.factor})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        grid = np.repeat(grid, self.factor, axis=0)
        return grid


class IncreaseWidth(Augmenter):
    def __init__(self, factor: int):
        assert factor > 1
        self.factor = factor

    def __str__(self):
        return f"IncreaseWidth({self.factor})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        del rng
        grid = np.repeat(grid, self.factor, axis=1)
        return grid


class DropoutInput(Augmenter):
    """
    Delete a random rectangular patch
    """

    def __init__(self):
        self.dropout_color = 0

    def __str__(self):
        return "DropoutInput()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        # get rng with seed
        if rng is None:
            rng = RandomState()
        grid = grid.copy()
        num_dropouts = rng.randint(1, 3)
        for _ in range(num_dropouts):
            x_len_ratio = rng.uniform(0.1, 0.3)
            y_len_ratio = rng.uniform(0.1, 0.3)
            x_start_ratio = rng.uniform(0.1, 0.7)
            y_start_ratio = rng.uniform(0.1, 0.7)

            x_len = int(np.ceil(grid.shape[0] * x_len_ratio))
            y_len = int(np.ceil(grid.shape[1] * y_len_ratio))

            x_start = int(grid.shape[0] * x_start_ratio)
            y_start = int(grid.shape[1] * y_start_ratio)

            grid[x_start : x_start + x_len, y_start : y_start + y_len] = (
                self.dropout_color
            )
        return grid

    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:
        # find unused colors in the task
        all_colors = set()
        for example in task.train_examples:
            all_colors.update(set(example.input.flatten().tolist()))
            all_colors.update(set(example.output.flatten().tolist()))

        all_colors = list(set(range(10)) - all_colors)
        if len(all_colors) == 0:
            self.dropout_color = 0
        else:
            self.dropout_color = rng.choice(all_colors)

        return super().apply_to_task(task, rng=rng, share_rng=share_rng, **kwargs)


class DropoutOutput(Augmenter):
    share_rng = True

    def __str__(self):
        return "DropoutOutput()"
        self.dropout_color = 0

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        assert rng is not None
        grid = grid.copy()
        num_dropouts = rng.randint(1, 3)
        for _ in range(num_dropouts):
            x_len_ratio = rng.uniform(0.1, 0.3)
            y_len_ratio = rng.uniform(0.1, 0.3)
            x_start_ratio = rng.uniform(0.1, 0.7)
            y_start_ratio = rng.uniform(0.1, 0.7)

            x_len = int(np.ceil(grid.shape[0] * x_len_ratio))
            y_len = int(np.ceil(grid.shape[1] * y_len_ratio))

            x_start = int(grid.shape[0] * x_start_ratio)
            y_start = int(grid.shape[1] * y_start_ratio)

            grid[x_start : x_start + x_len, y_start : y_start + y_len] = (
                self.dropout_color
            )
        return grid

    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:
        # find unused colors in the task
        all_colors = set()
        for example in task.train_examples:
            all_colors.update(set(example.input.flatten().tolist()))
            all_colors.update(set(example.output.flatten().tolist()))

        all_colors = list(set(range(10)) - all_colors)

        if len(all_colors) == 0:
            self.dropout_color = 0
        else:
            self.dropout_color = rng.choice(all_colors)

        return super().apply_to_task(task, rng=rng, share_rng=share_rng, **kwargs)


class RandomTranslateXY(Augmenter):
    share_rng = True

    def __str__(self):
        return "RandomTranslateXY()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        assert rng is not None
        # if rng.rand() < 0.5:
        shift_x = rng.randint(0, min(4, grid.shape[0]))
        shift_y = rng.randint(0, min(4, grid.shape[1]))
        # else:
        #     shift_x_ratio = rng.uniform(-0.5, 0.5)
        #     shift_x = int(np.round(grid.shape[0] * shift_x_ratio))
        #     shift_y_ratio = rng.uniform(-0.5, 0.5)
        #     shift_y = int(np.round(grid.shape[1] * shift_y_ratio))
        return np.roll(grid, (shift_x, shift_y), axis=(0, 1))


class RandomTranslateX(Augmenter):
    share_rng = True

    def __str__(self):
        return "RandomTranslateX()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        assert rng is not None
        # if rng.rand() < 0.5:
        shift_x = rng.randint(1, min(4, grid.shape[0]))
        # else:
        #     shift_x_ratio = rng.uniform(-0.5, 0.5)
        #     shift_x = int(np.round(grid.shape[0] * shift_x_ratio))
        return np.roll(grid, shift_x, axis=0)


class RandomTranslateY(Augmenter):
    share_rng = True

    def __str__(self):
        return "RandomTranslateY()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        assert rng is not None
        # if rng.rand() < 0.5:
        shift_y = rng.randint(1, min(4, grid.shape[1]))
        # else:
        #     shift_y_ratio = rng.uniform(-0.5, 0.5)
        #     shift_y = int(np.round(grid.shape[1] * shift_y_ratio))
        return np.roll(grid, shift_y, axis=1)


def find_connected_components(
    grid: Grid, background_color: Optional[int] = None
) -> List[List[tuple[int, int]]]:
    # get background color as the most frequenct color
    if background_color is None:
        background_color = np.bincount(grid.flatten()).argmax()

    # get connected components
    visited = np.zeros_like(grid)
    components = []
    for i in range(grid.shape[0]):
        for j in range(grid.shape[1]):
            if visited[i, j] == 0 and grid[i, j] != background_color:
                component = []
                stack = [(i, j)]
                while stack:
                    x, y = stack.pop()
                    if x < 0 or x >= grid.shape[0] or y < 0 or y >= grid.shape[1]:
                        continue
                    if visited[x, y] == 1 or grid[x, y] == background_color:
                        continue
                    component.append((x, y))
                    visited[x, y] = 1
                    stack.append((x + 1, y))
                    stack.append((x - 1, y))
                    stack.append((x, y + 1))
                    stack.append((x, y - 1))
                components.append(component)

    return components, background_color


class RandomObjectRotate(Augmenter):
    share_rng = True

    def __init__(self, angle: int):
        self.angle = angle

    def __str__(self):
        return f"RandomObjectRotate({self.angle})"

    def apply_to_grid(
        self, grid: Grid, rng: RandomState = None, background_color: int = 0
    ) -> Grid:
        assert rng is not None

        components, background_color = find_connected_components(
            grid, background_color=background_color
        )

        if not components:
            return grid

        idx = rng.choice(len(components), size=1)[0]
        component = components[idx]

        # copy grid
        grid = grid.copy()

        # rotate part of the grid by the given angle assume left bottom corner is the origin
        rotated_component = []
        origin = np.array([min(x for x, y in component), min(y for x, y in component)])
        for x, y in component:
            color = grid[x, y]
            x -= origin[0]
            y -= origin[1]
            if self.angle == 90:
                x, y = y, -x
            elif self.angle == 180:
                x, y = -x, -y
            elif self.angle == 270:
                x, y = -y, x
            rotated_component.append((x + origin[0], y + origin[1], color))

        for x, y in component:
            grid[x, y] = background_color
        for x, y, color in rotated_component:
            grid[x, y] = color

        return grid


class PermuteExamples(Augmenter):
    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:
        if rng is None:
            rng = RandomState()

        perm = rng.permutation(len(task.train_examples))
        train_examples = [task.train_examples[i] for i in perm]
        return Task(
            train_examples=train_examples,
            test_example=task.test_example,
            name=task.name,
        )


class RandomObjectTranslateXY(Augmenter):
    share_rng = True

    def __str__(self):
        return "RandomObjectTranslateXY()"

    def apply_to_grid(
        self,
        grid: Grid,
        rng: RandomState = None,
        background_color: Optional[int] = None,
    ) -> Grid:
        assert rng is not None

        components, background_color = find_connected_components(
            grid, background_color=background_color
        )

        if not components:
            return grid

        idx = rng.choice(len(components), size=1)[0]
        component = components[idx]

        # copy grid
        grid = grid.copy()

        # translate part of the grid by the given angle assume left bottom corner is the origin
        translated_component = []
        shift_x = rng.randint(-grid.shape[0] // 2, grid.shape[0] // 2)
        shift_y = rng.randint(-grid.shape[1] // 2, grid.shape[1] // 2)
        for x, y in component:
            color = grid[x, y]
            x += shift_x
            y += shift_y
            translated_component.append((x, y, color))

        for x, y in component:
            grid[x, y] = background_color
        for x, y, color in translated_component:
            grid[x, y] = color

        return grid


class Chain(Augmenter):
    def __init__(self, augmenters: Tuple[Augmenter]):
        self.augmenters = augmenters

    def __str__(self):
        return f"Chain({self.augmenters})"

    def apply_to_task(
        self, task: Task, rng: RandomState = None, share_rng=False, **kwargs
    ) -> Task:
        for augmenter in self.augmenters:
            task = augmenter.apply_to_task(
                task, rng=rng, share_rng=augmenter.share_rng, **kwargs
            )
        return task

    def apply_to_example(
        self,
        example: Example,
        rng: RandomState = None,
        to_input: bool = True,
        to_output: bool = True,
    ) -> Example:
        for augmenter in self.augmenters:
            example = augmenter.apply_to_example(
                example, rng=rng, to_input=to_input, to_output=to_output
            )
        return example

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        for augmenter in self.augmenters:
            grid = augmenter.apply_to_grid(grid, rng=rng)
        return grid


class Concat(Augmenter):
    def __init__(self, augmenters: Tuple[Augmenter], axis: int = 0):
        self.augmenters = augmenters
        self.axis = axis

    def __str__(self):
        return f"Concat({self.augmenters}, axis={self.axis})"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        grids = []
        for augmenter in self.augmenters:
            grids.append(augmenter.apply_to_grid(grid, rng=rng))

        return np.concatenate(grids, axis=self.axis)


class IdentityAugmenter(Augmenter):
    def __str__(self):
        return "IdentityAugmenter()"

    def apply_to_grid(self, grid: Grid, rng: RandomState = None) -> Grid:
        return grid.copy()


def inverse(augmenter):
    if isinstance(augmenter, Rotate):
        return Rotate(360 - augmenter.angle)
    elif isinstance(augmenter, Flip):
        return augmenter
    elif isinstance(augmenter, Transpose):
        return augmenter
    elif isinstance(augmenter, PermuteColors):
        color_map = augmenter._color_map
        # reverse
        inverse_map = {v: k for k, v in color_map.items()}
        return PermuteColorswithMap(inverse_map)


augmenters_to_apply_to_input = [
    DropoutInput(),
    IncreaseResolution(2),
    IncreaseResolution(3),
    IncreaseHeight(2),
    IncreaseWidth(2),
    IncreaseHeight(3),
    IncreaseWidth(3),
]

input_augmenters_probs = [
    1,
    1 / 2,
    1 / 2,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
]

augmenters_to_apply_to_output = [
    Rotate(90),
    Rotate(270),
    Rotate(180),
    Flip(0),
    Flip(1),
    Reflect(0, reverse=True),
    Reflect(1, reverse=True),
    Reflect(0, reverse=False),
    Reflect(1, reverse=False),
    IncreaseResolution(2),
    IncreaseResolution(3),
    IncreaseHeight(2),
    IncreaseWidth(2),
    Transpose(),
    RandomTranslateXY(),
    DropoutOutput(),
    Repeat(0, 2),
    Repeat(1, 2),
    Repeat(0, 3),
    Repeat(1, 3),
    Repeat(2, 2),
    Repeat(2, 3),
]

output_augmenters_probs = [
    1 / 3,
    1 / 3,
    1 / 3,
    1 / 2,
    1 / 2,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 2,
    1 / 2,
    1 / 2,
    1 / 5,
    1 / 5,
    1 / 5,
    1 / 5,
    1 / 5,
    1 / 5,
]

augmenters_to_apply_to_both = [
    IncreaseResolution(2),
    IncreaseResolution(3),
    IncreaseHeight(2),
    IncreaseWidth(2),
    IncreaseHeight(3),
    IncreaseWidth(3),
    PermuteColors(),
    # Transpose(),
    # Rotate(90),
    # Rotate(270),
    # Rotate(180),
    # Flip(0),
    # Flip(1),
    # Reflect(0, reverse=False),
    # Reflect(1, reverse=False),
]

both_augmenters_probs = [
    1 / 2,
    1 / 2,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 4,
    1 / 2,
    # 1/8,
    # 1/8,
    # 1/8,
    # 1/8,
    # 1/8,
    # 1/8,
    # 1/8,
    # 1/8,
]

# normalize
input_augmenters_probs = [
    p / sum(input_augmenters_probs) for p in input_augmenters_probs
]
output_augmenters_probs = [
    p / sum(output_augmenters_probs) for p in output_augmenters_probs
]
both_augmenters_probs = [p / sum(both_augmenters_probs) for p in both_augmenters_probs]


def apply_a_random_augmentation(task: Task, rng=None) -> Tuple[Task, Augmenter, str]:
    if rng is None:
        rng = RandomState()

    category = rng.choice(["input", "output", "both"], p=[0.3, 0.6, 0.1])

    if category == "input":
        augmenter = rng.choice(augmenters_to_apply_to_input, p=input_augmenters_probs)
        task = augmenter.apply_to_task(task, to_input=True, to_output=False, rng=rng)
    elif category == "output":
        augmenter = rng.choice(augmenters_to_apply_to_output, p=output_augmenters_probs)
        task = augmenter.apply_to_task(task, to_input=False, to_output=True, rng=rng)
    else:
        augmenter = rng.choice(augmenters_to_apply_to_both, p=both_augmenters_probs)
        task = augmenter.apply_to_task(task, to_input=True, to_output=True, rng=rng)

    return task, augmenter, category


if __name__ == "__main__":

    grid = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2, 2, 2],
            [3, 3, 3, 3, 3, 3, 3],
            [4, 4, 4, 4, 4, 4, 4],
            [5, 5, 5, 5, 5, 5, 5],
        ]
    )

    rng = RandomState(45)
    task = Task(train_examples=[Example(grid, grid)], test_example=Example(grid, grid))

    drop_task = DropoutOutput().apply_to_task(
        task, rng=rng, to_input=False, to_output=True
    )
    drop_task_input = DropoutInput().apply_to_task(
        task, rng=rng, to_input=True, to_output=False
    )

    grid1 = np.array(
        [
            [0, 0, 0, 0, 0],
            [0, 1, 1, 1, 0],
            [0, 1, 1, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 2, 2, 0, 0],
            [0, 2, 2, 0, 0],
            [0, 2, 2, 0, 0],
        ]
    )

    # another grid
    grid2 = np.array(
        [
            [0, 0, 0, 0, 0],
            [0, 5, 5, 5, 0],
            [0, 5, 5, 5, 0],
            [0, 0, 0, 0, 0],
            [0, 2, 4, 0, 0],
            [0, 9, 4, 0, 0],
            [0, 9, 4, 0, 0],
        ]
    )

    train_examples = [Example(grid1, grid1) for _ in range(3)]
    train_examples.append(Example(grid1, grid2))
    test_example = Example(grid1, grid1)

    task = Task(train_examples=train_examples, test_example=test_example)

    for _ in range(1000):
        ttask, augmenterr, category = apply_a_random_augmentation(task, rng=rng)
        assert np.array_equal(
            ttask.train_examples[0].output, ttask.train_examples[1].output
        )
        assert np.array_equal(
            ttask.train_examples[0].output, ttask.train_examples[2].output
        )
        assert not np.array_equal(
            ttask.train_examples[0].output, ttask.train_examples[3].output
        )

    # apply permute colors
    permute_colors = PermuteColors()
    ttask = permute_colors.apply_to_task(task, rng=rng)

    # random permutation
    permuted = rng.permutation(np.arange(10))

    print(permuted)

    def color_mapper(color: int) -> int:
        return permuted[color]

    color_mapper_v = np.vectorize(color_mapper)

    color_mapper_v(task.test_example.input)



from abc import abstractmethod
from typing import Tuple

import cvxopt
import cvxpy as cp
import numpy as np
import torch
from torch import Tensor

"""Implementation of Pareto HyperNetworks with:
1. Linear scalarization
3. EPO


EPO code from: https://github.com/dbmptr/EPOSearch
"""


class Solver:
    def __init__(self, n_tasks):
        super().__init__()
        self.n_tasks = n_tasks

    @abstractmethod
    def get_weighted_loss(self, losses, ray, parameters=None, **kwargs):
        pass

    def __call__(self, losses, ray, parameters, **kwargs):
        return self.get_weighted_loss(losses, ray, parameters, **kwargs)


class LinearScalarizationSolver(Solver):
    """For LS we use the preference ray to weigh the losses"""

    def __init__(self, n_tasks):
        super().__init__(n_tasks)

    def get_weighted_loss(self, losses, ray, parameters=None, **kwargs):
        return (losses * ray).sum()


class EPOSolver(Solver):
    """Wrapper over EPO"""

    def __init__(self, n_tasks, n_params):
        super().__init__(n_tasks)
        self.solver = EPO(n_tasks=n_tasks, n_params=n_params)

    def get_weighted_loss(
        self, losses, ray, parameters: Tuple[Tensor] = None, **kwargs
    ):
        assert parameters is not None
        return self.solver.get_weighted_loss(losses, ray, parameters)


class EPO:
    def __init__(self, n_tasks, n_params):
        self.n_tasks = n_tasks
        self.n_params = n_params

    def __call__(self, losses, ray, parameters):
        return self.get_weighted_loss(losses, ray, parameters)

    @staticmethod
    def _flattening(grad: Tuple[Tensor]) -> Tensor:
        return torch.cat(tuple(g.flatten() for g in grad), dim=0)

    def get_weighted_loss(
        self,
        losses: Tuple[Tensor],
        ray: Tensor,
        parameters: Tuple[Tensor],
    ):
        lp = ExactParetoLP(m=self.n_tasks, n=self.n_params, r=ray.cpu().numpy())

        grads = []
        for i, loss in enumerate(losses):
            g = torch.autograd.grad(
                loss, parameters, retain_graph=True, allow_unused=True
            )
            g = tuple(filter(lambda x: x is not None, g))
            flat_grad = self._flattening(g)
            grads.append(flat_grad.data)

        G = torch.stack(grads)
        GG_T = G @ G.T
        GG_T = GG_T.detach().cpu().numpy()

        numpy_losses = losses.detach().cpu().numpy()

        try:
            alpha = lp.get_alpha(numpy_losses, G=GG_T, C=True)
        except Exception as excep:
            print(excep)
            alpha = None

        if alpha is None:  # A patch for the issue in cvxpy
            alpha = (ray / ray.sum()).cpu().numpy()

        alpha *= self.n_tasks
        alpha = torch.from_numpy(alpha).to(losses.device)

        weighted_loss = torch.sum(losses * alpha)
        return weighted_loss


class ExactParetoLP(object):
    """modifications of the code in https://github.com/dbmptr/EPOSearch"""

    def __init__(self, m, n, r, eps=1e-4):
        cvxopt.glpk.options["msg_lev"] = "GLP_MSG_OFF"
        self.m = m
        self.n = n
        self.r = r
        self.eps = eps
        self.last_move = None
        self.a = cp.Parameter(m)  # Adjustments
        self.C = cp.Parameter((m, m))  # C: Gradient inner products, G^T G
        self.Ca = cp.Parameter(m)  # d_bal^TG
        self.rhs = cp.Parameter(m)  # RHS of constraints for balancing

        self.alpha = cp.Variable(m)  # Variable to optimize

        obj_bal = cp.Maximize(self.alpha @ self.Ca)  # objective for balance
        constraints_bal = [
            self.alpha >= 0,
            cp.sum(self.alpha) == 1,  # Simplex
            self.C @ self.alpha >= self.rhs,
        ]
        self.prob_bal = cp.Problem(obj_bal, constraints_bal)  # LP balance

        obj_dom = cp.Maximize(cp.sum(self.alpha @ self.C))  # obj for descent
        constraints_res = [
            self.alpha >= 0,
            cp.sum(self.alpha) == 1,  # Restrict
            self.alpha @ self.Ca >= -cp.neg(cp.max(self.Ca)),
            self.C @ self.alpha >= 0,
        ]
        constraints_rel = [
            self.alpha >= 0,
            cp.sum(self.alpha) == 1,  # Relaxed
            self.C @ self.alpha >= 0,
        ]
        self.prob_dom = cp.Problem(obj_dom, constraints_res)  # LP dominance
        self.prob_rel = cp.Problem(obj_dom, constraints_rel)  # LP dominance

        self.gamma = 0  # Stores the latest Optimum value of the LP problem
        self.mu_rl = 0  # Stores the latest non-uniformity

    def get_alpha(self, loss_terms, G: Tensor, r=None, C: bool = False, relax=False):
        r = self.r if r is None else r
        assert len(loss_terms) == len(G) == len(r) == self.m, "length != m"
        rl, self.mu_rl, self.a.value = adjustments(loss_terms, r)
        self.C.value = G if C else G @ G.T
        self.Ca.value = self.C.value @ self.a.value

        if self.mu_rl > self.eps:
            J = self.Ca.value > 0
            if len(np.where(J)[0]) > 0:
                J_star_idx = np.where(rl == np.max(rl))[0]
                self.rhs.value = self.Ca.value.copy()
                self.rhs.value[J] = -np.inf  # Not efficient; but works.
                self.rhs.value[J_star_idx] = 0
            else:
                self.rhs.value = np.zeros_like(self.Ca.value)
            self.gamma = self.prob_bal.solve(solver=cp.GLPK, verbose=False)
            self.last_move = "bal"
        else:
            if relax:
                self.gamma = self.prob_rel.solve(solver=cp.GLPK, verbose=False)
            else:
                self.gamma = self.prob_dom.solve(solver=cp.GLPK, verbose=False)
            self.last_move = "dom"

        return self.alpha.value


def mu(rl, normed=False):
    if len(np.where(rl < 0)[0]):
        raise ValueError(f"rl<0 \n rl={rl}")
        # return None
    m = len(rl)
    l_hat = rl if normed else rl / rl.sum()
    eps = np.finfo(rl.dtype).eps
    l_hat = l_hat[l_hat > eps]
    return np.sum(l_hat * np.log(l_hat * m))


def adjustments(l, r=1):
    m = len(l)
    rl = r * l
    l_hat = rl / rl.sum()
    mu_rl = mu(l_hat, normed=True)
    a = r * (np.log(l_hat * m) - mu_rl)
    return rl, mu_rl, a



import contextlib
from typing import Dict, Generator, Iterable, Optional, Tuple

import torch
from transformers.utils import (
    is_torch_bf16_gpu_available,
    is_torch_cuda_available,
    is_torch_mps_available,
    is_torch_npu_available,
    is_torch_xpu_available,
)

PRECISION_STR_TO_DTYPE: Dict[str, torch.dtype] = {
    "fp16": torch.float16,
    "float16": torch.float16,
    "bf16": torch.bfloat16,
    "bfloat16": torch.bfloat16,
    "float": torch.float32,
    "fp32": torch.float32,
    "float32": torch.float32,
    "double": torch.float64,
    "fp64": torch.float64,
    "float64": torch.float64,
}


def parse_dtype(dtype: Optional[str]):
    """
    Parses a string representation of a data type and returns the corresponding torch.dtype.

    Args:
        dtype (Optional[str]): The string representation of the data type.
                               Can be one of "float32", "float", "float64", "double",
                               "float16", "half", "bfloat16", or "bf16".
                               If None, returns None.

    Returns:
        torch.dtype: The corresponding torch.dtype if the input is a valid string representation.
                     If the input is already a torch.dtype, it is returned as is.
                     If the input is None, returns None.

    Raises:
        ValueError: If the input string does not correspond to a supported data type.
    """
    if isinstance(dtype, torch.dtype):
        return dtype

    if dtype is None:
        return None

    dtype = dtype.strip('"')
    if dtype not in PRECISION_STR_TO_DTYPE:
        raise ValueError(f"Unsupported dtype: {type(dtype)}")

    dtype = PRECISION_STR_TO_DTYPE[dtype]
    return dtype


def get_dtype(obj) -> torch.dtype:
    """
    Get the data type (dtype) of a given object.

    Returns:
        torch.dtype: The data type of the given object.

    Raises:
        ValueError: If the object type is not supported.
    """
    if isinstance(obj, torch.Tensor):
        return obj.dtype
    elif isinstance(obj, torch.nn.Module):
        if hasattr(obj, "dtype"):
            return obj.dtype
        else:
            return next(iter(obj.parameters())).dtype
    elif isinstance(obj, (torch.device, str)):
        return parse_dtype(obj)
    else:
        raise ValueError(f"Unsupported object type: {type(obj)}")


@contextlib.contextmanager
def set_default_dtype(dtype: torch.dtype) -> Generator[None, None, None]:
    """
    Context manager to set torch's default dtype.

    Args:
        dtype (torch.dtype): The desired default dtype inside the context manager.

    Returns:
        ContextManager: context manager for setting default dtype.

    Example:
        >>> with set_default_dtype(torch.bfloat16):
        >>>     x = torch.tensor([1, 2, 3])
        >>>     x.dtype
        torch.bfloat16


    """
    old_dtype = torch.get_default_dtype()
    torch.set_default_dtype(dtype)
    try:
        yield
    finally:
        torch.set_default_dtype(old_dtype)


def infer_optim_dtype(model_dtype: "torch.dtype") -> "torch.dtype":
    r"""
    Infers the optimal dtype according to the model_dtype and device compatibility.
    """
    _is_fp16_available = is_torch_npu_available() or is_torch_cuda_available()
    try:
        _is_bf16_available = is_torch_bf16_gpu_available() or (
            is_torch_npu_available() and torch.npu.is_bf16_supported()
        )
    except Exception:
        _is_bf16_available = False

    if _is_bf16_available and model_dtype == torch.bfloat16:
        return torch.bfloat16
    elif _is_fp16_available:
        return torch.float16
    else:
        return torch.float32


def validate_expected_param_dtype(
    named_params: Iterable[Tuple[str, torch.nn.Parameter]], dtype: torch.dtype
) -> None:
    """
    Validates that all input parameters have the expected dtype.

    Args:
        named_params (Iterable[Tuple[str, torch.nn.Parameter]]): Iterable of named parameters.
        dtype (torch.dtype): Expected dtype.

    Raises:
        ValueError: If any parameter has a different dtype than `dtype`.
    """
    for name, param in named_params:
        if param.dtype != dtype:
            raise ValueError(
                f"Parameter {name} has dtype {param.dtype}, but expected {dtype}"
            )



import torch
from torch import Tensor, nn

from fusion_bench import BaseAlgorithm, BaseModelPool
from fusion_bench.utils.state_dict_arithmetic import state_dict_sum

from .utils import (
    module_random_drop_,
    module_sub_,
    param_random_drop_,
    trainable_state_dict,
)


class DareTaskArithmetic(BaseAlgorithm):
    """
    Implementation of Task Arithmetic w/ DARE.

    - Yu et al. Language Models are Super Mario: Absorbing Abilities from Homologous Models as a Free Lunch. 2023. http://arxiv.org/abs/2311.03099
    """

    def __init__(
        self,
        scaling_factor: float,
        sparsity_ratio: float,
        only_on_linear_weights: bool,
        rescale: bool = True,
        **kwargs,
    ):
        self.scaling_factor = scaling_factor
        self.sparsity_ratio = sparsity_ratio
        self.only_on_linear_weights = only_on_linear_weights
        self.rescale = rescale
        super().__init__(**kwargs)

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        assert (
            self.sparsity_ratio >= 0 and self.sparsity_ratio <= 1
        ), "Sparsity ratio must be between 0 and 1"
        pretrained_model = modelpool.load_pretrained_model()
        finetuned_models = {
            model_name: modelpool.load_model(model_name)
            for model_name in modelpool.model_names
        }
        task_vectors = {
            model_name: module_sub_(finetuned_models[model_name], pretrained_model)
            for model_name in finetuned_models
        }
        del finetuned_models

        # drop and rescale task vectors
        for model_name, tv in task_vectors.items():
            if self.only_on_linear_weights:
                for module_name, module in tv.named_modules():
                    if isinstance(module, nn.Linear):
                        print(f"pruning model: `{model_name}`, layer: {module_name}.")
                        param_random_drop_(
                            module.weight, self.sparsity_ratio, rescale=self.rescale
                        )
            else:
                print(f"pruning model: `{model_name}`")
                module_random_drop_(tv, self.sparsity_ratio, rescale=self.rescale)

        # merge task vectors
        task_vector_sum = state_dict_sum(
            [trainable_state_dict(tv) for tv in task_vectors.values()]
        )

        # scale the task vector and add it to the pretrained model
        for name, delta in task_vector_sum.items():
            delta = delta * self.scaling_factor
            pretrained_model.get_parameter(name).data.add_(delta)

        return pretrained_model



# flake8: noqa F401
from .clip_rankone_moe_taskpool import RankoneMoECLIPVisionModelTaskPool
from .clip_sparse_wemoe_taskpool import SparseWEMoECLIPVisionModelTaskPool
from .taskpool import CLIPVisionModelTaskPool



import math
import time

import torch
import torch.nn as nn
import transformers

torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.allow_tf32 = False


class AblateGPT:
    def __init__(self, layer):
        self.layer = layer
        self.dev = self.layer.weight.device
        W = layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        self.rows = W.shape[0]
        self.columns = W.shape[1]
        self.H = torch.zeros((self.columns, self.columns), device=self.dev)
        self.nsamples = 0

        self.scaler_row = torch.zeros((self.columns), device=self.dev)

    def add_batch(self, inp, out):
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or isinstance(
            self.layer, transformers.Conv1D
        ):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        self.H *= self.nsamples / (self.nsamples + tmp)

        self.scaler_row *= self.nsamples / (self.nsamples + tmp)

        self.nsamples += tmp
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        self.H += inp.matmul(inp.t())
        self.scaler_row += torch.norm(inp, p=2, dim=1) ** 2 / self.nsamples

    def get_wanda_mask(self, sparsity, prunen, prunem):
        W_metric = torch.abs(self.layer.weight.data) * torch.sqrt(
            self.scaler_row.reshape((1, -1))
        )
        W_mask = torch.zeros_like(W_metric) == 1  ## initialize a mask to be all False
        if prunen != 0:
            for ii in range(W_metric.shape[1]):
                if ii % prunem == 0:
                    tmp = W_metric[:, ii : (ii + prunem)].float()
                    W_mask.scatter_(
                        1, ii + torch.topk(tmp, prunen, dim=1, largest=False)[1], True
                    )
        else:
            sort_res = torch.sort(W_metric, dim=-1, stable=True)
            indices = sort_res[1][:, : int(W_metric.shape[1] * sparsity)]
            W_mask.scatter_(1, indices, True)

        return W_mask

    def get_mag_mask(self, sparsity, prunen, prunem):
        W = self.layer.weight.data
        W_metric = torch.abs(W)
        if prunen != 0:
            W_mask = torch.zeros_like(W) == 1
            for ii in range(W_metric.shape[1]):
                if ii % prunem == 0:
                    tmp = W_metric[:, ii : (ii + prunem)].float()
                    W_mask.scatter_(
                        1, ii + torch.topk(tmp, prunen, dim=1, largest=False)[1], True
                    )
        else:
            thresh = torch.sort(W_metric.flatten().cuda())[0][
                int(W.numel() * sparsity)
            ].cpu()
            W_mask = W_metric <= thresh

        return W_mask

    def fasterprune(
        self,
        args,
        sparsity,
        mask=None,
        prune_n=0,
        prune_m=0,
        blocksize=128,
        percdamp=0.01,
    ):
        W = self.layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        H = self.H
        del self.H
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0

        Losses = torch.zeros(self.rows, device=self.dev)

        damp = percdamp * torch.mean(torch.diag(H))
        diag = torch.arange(self.columns, device=self.dev)
        H[diag, diag] += damp
        H = torch.linalg.cholesky(H)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            if prune_n == 0 or mask is not None:
                if mask is not None:
                    mask1 = mask[:, i1:i2]
                else:
                    # tmp = W1 ** 2 / (torch.diag(Hinv1).reshape((1, -1))) ** 2
                    if "wanda" in args.prune_method:
                        tmp = torch.abs(W1) * torch.sqrt(
                            self.scaler_row[i1:i2].reshape((1, -1))
                        )
                    elif "mag" in args.prune_method:
                        tmp = torch.abs(W1)
                    thresh = torch.sort(tmp.flatten())[0][int(tmp.numel() * sparsity)]
                    mask1 = tmp <= thresh
            else:
                mask1 = torch.zeros_like(W1) == 1

            for i in range(count):
                w = W1[:, i]
                d = Hinv1[i, i]

                if prune_n != 0 and i % prune_m == 0 and mask is None:
                    # tmp = W1[:, i:(i + prune_m)] ** 2 / (torch.diag(Hinv1)[i:(i + prune_m)].reshape((1, -1))) ** 2
                    if "wanda" in args.prune_method:
                        tmp = torch.abs(W1[:, i : (i + prune_m)]) * torch.sqrt(
                            self.scaler_row[(i + i1) : (i + i1 + prune_m)].reshape(
                                (1, -1)
                            )
                        )
                    elif "mag" in args.prune_method:
                        tmp = torch.abs(W1[:, i : (i + prune_m)])
                    mask1.scatter_(
                        1, i + torch.topk(tmp, prune_n, dim=1, largest=False)[1], True
                    )

                q = w.clone()
                q[mask1[:, i]] = 0

                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1

            W[:, i1:i2] = Q1
            Losses += torch.sum(Losses1, 1) / 2

            W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])

        torch.cuda.synchronize()
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        self.layer.weight.data = W.reshape(self.layer.weight.shape).to(
            self.layer.weight.data.dtype
        )

    def free(self):
        self.H = None
        torch.cuda.empty_cache()



import functools
import logging
import os
from copy import deepcopy

import torch
from torch import Tensor
from torch.utils.data import DataLoader
from transformers.models.clip.modeling_clip import CLIPEncoder

from fusion_bench.dataset import CLIPDataset
from fusion_bench.method.task_arithmetic.task_arithmetic import task_arithmetic_merge
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.models.rankone_moe import RankOneMoE
from fusion_bench.utils.data import InfiniteDataLoader

from .rankone_moe import RankOneMoEAlgorithm

log = logging.getLogger(__name__)


class CLIPRankOneMoEAlgorithm(
    RankOneMoEAlgorithm,
    CLIPClassificationMixin,
):
    """
    CLIPRankOneMoEAlgorithm is a class that implements the RankOneMoEAlgorithm (https://github.com/EnnengYang/RankOne-MoE)
    for CLIP models. It extends the RankOneMoEAlgorithm and CLIPClassificationMixin classes.

    Attributes:
        modelpool (CLIPVisionModelPool): The model pool containing the CLIP models.
    """

    modelpool: CLIPVisionModelPool = None

    def load_checkpoint(self, model, checkpoint):
        """
        Load the checkpoint file.

        Args:
            model: The model to load the checkpoint into.
            checkpoint: The path to the checkpoint file.
        """
        state = {"model": model}
        self._fabric.load(checkpoint, state)

    def save_checkpoint(self, model, checkpoint):
        """
        Save the checkpoint file.

        Args:
            model: The model to save the checkpoint from.
            checkpoint: The path to the checkpoint file.
        """
        self._fabric.save(checkpoint, {"model": model})

    def construct_moe_model(self) -> RankOneMoE:
        """
        Construct the RankOne-MoE model using the models in the model pool.

        Returns:
            RankOne-MoE: The constructed MoE model.
        """
        base_model = self.modelpool.load_model("_pretrained_")
        expert_models = [
            self.modelpool.load_model(m) for m in self.modelpool.model_names
        ]

        # Merge the models using task arithmetic
        moe_model = task_arithmetic_merge(
            # This function modifies the model in place, so we need to pass a deepcopy
            deepcopy(base_model),
            expert_models,
            scaling_factor=self.config.init_lambda,
        ).requires_grad_(False)

        # Up-scale MLP modules
        base_encoder: CLIPEncoder = base_model.vision_model.encoder
        moe_encoder: CLIPEncoder = moe_model.vision_model.encoder
        expert_encoders = [m.vision_model.encoder for m in expert_models]

        num_layers = len(base_encoder.layers)
        for layer_idx in range(num_layers):
            base_mlp = base_encoder.layers[layer_idx].mlp
            expert_mlps = [e.layers[layer_idx].mlp for e in expert_encoders]

            moe_encoder.layers[layer_idx].mlp = RankOneMoE(
                hidden_size=base_encoder.config.hidden_size,
                base_model=base_mlp,
                expert_models=expert_mlps,
                init_lambda=self.config.init_lambda,
                batch_first=True,  # For open_clip models this is False
                router_hidden_layers=self.config.router_hidden_layers,
                batch_reduce=self.config.batch_reduce,
                svd_accelerator=self.config.svd_accelerator,
                rank_k=self.config.rank_k,
                select_k=self.config.select_k,
            )

        return moe_model

    @functools.cache
    def get_shuffled_test_loader_iter(self, tta_dataset: str):
        """
        Get an iterator for the shuffled test data loader.

        Args:
            tta_dataset (str): The name of the test-time adaptation dataset.

        Returns:
            Iterator: An iterator for the shuffled test data loader.
        """
        dataset = self.modelpool.load_test_dataset(tta_dataset)
        dataset = CLIPDataset(dataset, processor=self.clip_processor)
        log.info("get_shuffled_test_loader_iter")
        loader = DataLoader(
            dataset,
            batch_size=self.config.batch_size,
            shuffle=True,
            num_workers=self.config.num_workers,
            pin_memory=True,
        )
        loader = self.fabric.setup_dataloaders(loader)
        return iter(InfiniteDataLoader(loader))

    def on_test_time_adaptation_start(self):
        """
        Load the CLIP processor and construct the zero-shot classification head for each task.
        """
        self.setup_zero_shot_classification_head()

    def compute_logits(self, module, batch, task) -> Tensor:
        """
        Compute the logits for the given batch and task.

        Args:
            module: The model module.
            batch: The input batch.
            task: The task name.

        Returns:
            Tensor: The computed logits.
        """
        images, _ = batch
        text_embeds = self.zeroshot_weights[task]

        image_embeds = module(images)[1]
        image_embeds = self.visual_projection(image_embeds)

        # Normalize embeddings
        image_embeds = image_embeds / image_embeds.norm(p=2, dim=-1, keepdim=True)

        # Cosine similarity
        logits_per_text = (
            torch.matmul(text_embeds, image_embeds.t()) * self.logit_scale_exp
        )
        logits_per_image = logits_per_text.t()

        return logits_per_image



"""
In this script, two reward functions are defined:

- compressibility, in which the file size of the image after JPEG compression is minimized
- incompressibility, in which the same measure is maximized.
"""

import io
from typing import List

import numpy as np
import torch
from PIL import Image


def jpeg_incompressibility_scorer():
    """
    Function to calculate the incompressibility score of an image.
    The score is calculated based on the size of the image after JPEG compression.
    The larger the size, the higher the incompressibility score.
    """

    def _fn(images: torch.Tensor, prompts, metadata):
        if isinstance(images, torch.Tensor):
            images = (images * 255).round().clamp(0, 255).to(torch.uint8).cpu().numpy()
            images = images.permute(0, 2, 3, 1)  # NCHW -> NHWC
        images: List[Image.Image] = [Image.fromarray(image) for image in images]
        buffers = [io.BytesIO() for _ in images]
        for image, buffer in zip(images, buffers):
            image.save(buffer, format="JPEG", quality=95)
        sizes = [buffer.tell() / 1000 for buffer in buffers]
        return torch.asarray(sizes), {}

    return _fn


def jpeg_compressibility_scorer():
    """
    Function to calculate the compressibility score of an image.
    The score is calculated based on the size of the image after JPEG compression.
    The smaller the size, the higher the compressibility score.
    """
    jpeg_fn = jpeg_incompressibility_scorer()

    def _fn(images: torch.Tensor, prompts, metadata):
        reward, metadata = jpeg_fn(images, prompts, metadata)
        return -reward, metadata

    return _fn



"""
This module provides a class to convert a dataset whose object is a list of dictionaries with keys "image" and "label" to a dataset whose object is a tuple of tensors (inputs, label) for CLIP models.
"""

from typing import Optional

import torch
from transformers import CLIPProcessor, ProcessorMixin


class CLIPDataset(torch.utils.data.Dataset):
    """
    A dataset class for CLIP models that converts a dataset of dictionaries or tuples
    into a format suitable for CLIP processing.

    This class wraps an existing dataset and applies CLIP preprocessing to the images.
    It expects each item in the dataset to be either a dictionary with 'image' and 'label' keys,
    or a tuple/list of (image, label).

    Args:
        dataset: The original dataset to wrap.
        processor (CLIPProcessor): The CLIP processor for preparing inputs. If None, no preprocessing is applied and raw images are returned.

    Attributes:
        dataset: The wrapped dataset.
        processor (CLIPProcessor): The CLIP processor used for image preprocessing.
    """

    def __init__(self, dataset, processor: Optional[CLIPProcessor] = None):
        self.dataset = dataset
        self.processor = processor

    def __len__(self):
        """Returns the number of items in the dataset."""
        return len(self.dataset)

    def __getitem__(self, idx: int):
        """
        Retrieves and processes an item from the dataset.

        Args:
            idx (int): The index of the item to retrieve.

        Returns:
            tuple: A tuple containing the processed image tensor and the label.

        Raises:
            ValueError: If the item is neither a dictionary nor a tuple/list of length 2.
        """
        item = self.dataset[idx]
        if isinstance(item, dict):
            item = item
        elif isinstance(item, (tuple, list)):
            assert len(item) == 2, "Each item should be a tuple or list of length 2"
            item = {"image": item[0], "label": item[1]}
        else:
            raise ValueError("Each item should be a dictionary or a tuple of length 2")
        image = item["image"]
        if self.processor is not None:
            if isinstance(self.processor, ProcessorMixin):
                # Apply the processor to the image to get the input tensor
                inputs = self.processor(images=[image], return_tensors="pt")[
                    "pixel_values"
                ][0]
        else:
            # if processor is None, return the raw image directly
            inputs = image
        return inputs, item["label"]



import logging
import time

log = logging.getLogger(__name__)


class timeit_context:
    """
    Usage:

    ```python
    with timeit_context() as timer:
        ... # code block to be measured
    ```
    """

    nest_level = -1

    def _log(self, msg):
        log.log(self.loglevel, msg, stacklevel=3)

    def __init__(self, msg: str = None, loglevel=logging.INFO) -> None:
        self.loglevel = loglevel
        self.msg = msg

    def __enter__(self) -> None:
        """
        Sets the start time and logs an optional message indicating the start of the code block execution.

        Args:
            msg: str, optional message to log
        """
        self.start_time = time.time()
        timeit_context.nest_level += 1
        if self.msg is not None:
            self._log("  " * timeit_context.nest_level + "[BEGIN] " + str(self.msg))

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Calculates the elapsed time and logs it, along with an optional message indicating the end of the code block execution.
        """
        end_time = time.time()
        elapsed_time = end_time - self.start_time
        self._log(
            "  " * timeit_context.nest_level
            + "[END]   "
            + str(f"Elapsed time: {elapsed_time:.2f}s")
        )
        timeit_context.nest_level -= 1
